from sonolus.backend.optimize.allocate import Allocate, AllocateBasic, AllocateFast
from sonolus.backend.optimize.constant_evaluation import SparseConditionalConstantPropagation
from sonolus.backend.optimize.copy_coalesce import CopyCoalesce
from sonolus.backend.optimize.dead_code import (
    AdvancedDeadCodeElimination,
    DeadCodeElimination,
    UnreachableCodeElimination,
)
from sonolus.backend.optimize.inlining import InlineVars
from sonolus.backend.optimize.simplify import (
    CoalesceFlow,
    CoalesceSmallConditionalBlocks,
    CombineExitBlocks,
    NormalizeSwitch,
    RemoveRedundantArguments,
    RewriteToSwitch,
)
from sonolus.backend.optimize.ssa import FromSSA, ToSSA

MINIMAL_PASSES = (
    CoalesceFlow(),
    UnreachableCodeElimination(),
    AllocateBasic(),
)

FAST_PASSES = (
    CoalesceFlow(),
    UnreachableCodeElimination(),
    AllocateFast(),  # Does dead code elimination too, so no need for a separate pass
    CoalesceFlow(),
)

STANDARD_PASSES = (
    CoalesceFlow(),
    UnreachableCodeElimination(),
    DeadCodeElimination(),
    CoalesceSmallConditionalBlocks(),
    ToSSA(),
    SparseConditionalConstantPropagation(),
    UnreachableCodeElimination(),
    DeadCodeElimination(),
    CoalesceFlow(),
    InlineVars(),
    DeadCodeElimination(),
    InlineVars(),
    CoalesceFlow(),
    SparseConditionalConstantPropagation(),
    RemoveRedundantArguments(),
    DeadCodeElimination(),
    CoalesceFlow(),
    RewriteToSwitch(),
    InlineVars(),
    FromSSA(),
    CoalesceFlow(),
    CopyCoalesce(),
    AdvancedDeadCodeElimination(),
    CoalesceFlow(),
    NormalizeSwitch(),
    CombineExitBlocks(),
    Allocate(),
)



from types import FunctionType
from typing import Any, Never, assert_never

from sonolus.backend.ops import Op
from sonolus.script.array import Array
from sonolus.script.array_like import ArrayLike
from sonolus.script.debug import error, require
from sonolus.script.internal.context import ctx
from sonolus.script.internal.dict_impl import DictImpl
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.math_impls import MATH_BUILTIN_IMPLS, _trunc
from sonolus.script.internal.native import native_function
from sonolus.script.internal.random import RANDOM_BUILTIN_IMPLS
from sonolus.script.internal.range import Range
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.internal.value import Value
from sonolus.script.iterator import (
    SonolusIterator,
    _EmptyIterator,
    _Enumerator,
    _FilteringIterator,
    _MappingIterator,
    _Zipper,
)
from sonolus.script.num import Num, _is_num
from sonolus.script.record import Record

_empty = object()


@meta_fn
def _isinstance(value, type_):
    value = validate_value(value)
    type_ = validate_value(type_)._as_py_()
    if type_ is dict:
        return isinstance(value, DictImpl)
    if type_ is tuple:
        return isinstance(value, TupleImpl)
    if type_ in {_int, _float, _bool}:
        raise TypeError("Instance check against int, float, or bool is not supported, use Num instead")
    if not (isinstance(type_, type) and (issubclass(type_, Value) or getattr(type_, "_allow_instance_check_", False))):
        raise TypeError(f"Unsupported type: {type_} for isinstance")
    return validate_value(isinstance(value, type_))


@meta_fn
def _len(value):
    from sonolus.backend.visitor import compile_and_call

    value = validate_value(value)
    if not hasattr(value, "__len__"):
        raise TypeError(f"object of type '{type(value).__name__}' has no len()")
    return compile_and_call(value.__len__)  # type: ignore


@meta_fn
def _enumerate(iterable, start=0):
    from sonolus.backend.visitor import compile_and_call

    iterable = validate_value(iterable)
    if isinstance(iterable, TupleImpl):
        return TupleImpl._accept_(tuple((start + i, value) for i, value in enumerate(iterable._as_py_(), start=start)))
    elif not hasattr(iterable, "__iter__"):
        raise TypeError(f"'{type(iterable).__name__}' object is not iterable")
    elif isinstance(iterable, ArrayLike):
        return compile_and_call(iterable._enumerate_, start)
    else:
        iterator = compile_and_call(iterable.__iter__)  # type: ignore
        if not isinstance(iterator, SonolusIterator):
            raise TypeError("Only subclasses of SonolusIterator are supported as iterators")
        return _Enumerator(0, start, iterator)


@meta_fn
def _reversed(iterable):
    from sonolus.backend.visitor import compile_and_call

    iterable = validate_value(iterable)
    if not isinstance(iterable, ArrayLike):
        raise TypeError(f"Unsupported type: {type(iterable)} for reversed")
    return compile_and_call(iterable.__reversed__)


@meta_fn
def _zip(*iterables, strict: bool = False):
    from sonolus.backend.visitor import compile_and_call
    from sonolus.script.containers import Pair

    if validate_value(strict)._as_py_():  # type: ignore
        raise NotImplementedError("Strict zipping is not supported")

    if not iterables:
        return _EmptyIterator()

    iterables = [validate_value(iterable) for iterable in iterables]
    if any(isinstance(iterable, TupleImpl) for iterable in iterables):
        if not all(isinstance(iterable, TupleImpl) for iterable in iterables):
            raise TypeError("Cannot mix tuples with other types in zip")
        return TupleImpl._accept_(tuple(zip(*(iterable.value for iterable in iterables), strict=False)))
    iterators = [compile_and_call(iterable.__iter__) for iterable in iterables]
    if not all(isinstance(iterator, SonolusIterator) for iterator in iterators):
        raise TypeError("Only subclasses of SonolusIterator are supported as iterators")
    v = iterators.pop()
    while iterators:
        v = Pair(iterators.pop(), v)
    return _Zipper(v)


@meta_fn
def _abs(value):
    from sonolus.backend.visitor import compile_and_call

    value = validate_value(value)
    if not hasattr(value, "__abs__"):
        raise TypeError(f"bad operand type for abs(): '{type(value).__name__}'")
    return compile_and_call(value.__abs__)  # type: ignore


def _identity(value):
    return value


@meta_fn
def _max(*args, default=_empty, key=None):
    from sonolus.backend.visitor import compile_and_call

    if key is None:
        key = _identity

    args = tuple(validate_value(arg) for arg in args)
    if len(args) == 0:
        raise ValueError("Expected at least one argument to max")
    elif len(args) == 1:
        (iterable,) = args
        if isinstance(iterable, ArrayLike):
            return compile_and_call(iterable._max_, key=key)
        elif isinstance(iterable, TupleImpl) and all(_is_num(v) for v in iterable.value):
            if len(iterable.value) == 0:
                if default is not _empty:
                    return default
                raise ValueError("max() arg is an empty sequence")
            return compile_and_call(Array(*iterable.value)._max_, key=key)
        elif isinstance(iterable, SonolusIterator):
            if not (default is _empty or Num._accepts_(default)):
                raise TypeError("default argument must be a number")
            return compile_and_call(
                _max_num_iterator,
                iterable,
                Num._accept_(default) if default is not _empty else None,
                key=key if key is not _identity else None,
            )
        else:
            raise TypeError(f"Unsupported type: {type(iterable)} for max")
    else:
        if default is not _empty:
            raise TypeError("default argument is not supported for max with multiple arguments")
        if not all(_is_num(arg) for arg in args):
            raise TypeError("Arguments to max must be numbers")
        if ctx():
            result = _max2(args[0], args[1], key=key)
            for arg in args[2:]:
                result = _max2(result, arg, key=key)
            return result
        else:
            return max(arg._as_py_() for arg in args)


def _max2(a, b, key=_identity):
    from sonolus.backend.visitor import compile_and_call

    a = validate_value(a)
    b = validate_value(b)
    if _is_num(a) and _is_num(b) and key == _identity:
        return compile_and_call(_max2_num, a, b)
    return compile_and_call(_max2_generic, a, b, key=key)


@native_function(Op.Max)
def _max2_num(a, b):
    if a > b:
        return a
    else:
        return b


def _max2_generic(a, b, key=_identity):
    if key(a) > key(b):
        return a
    else:
        return b


def _max_num_iterator(iterable, default, key):
    iterator = iterable.__iter__()  # noqa: PLC2801
    initial = iterator.next()
    if initial.is_nothing:
        require(default is not None, "default must be provided if the iterator is empty")
        return default
    if key is not None:
        result = initial.get_unsafe()
        best_key = key(result)
        for value in iterator:
            new_key = key(value)
            if new_key > best_key:
                result = value
                best_key = new_key
        return result
    else:
        result = initial.get_unsafe()
        for value in iterator:
            if value > result:  # noqa: PLR1730
                result = value
        return result


@meta_fn
def _min(*args, default=_empty, key=None):
    from sonolus.backend.visitor import compile_and_call

    if key is None:
        key = _identity

    args = tuple(validate_value(arg) for arg in args)
    if len(args) == 0:
        raise ValueError("Expected at least one argument to min")
    elif len(args) == 1:
        (iterable,) = args
        if isinstance(iterable, ArrayLike):
            return compile_and_call(iterable._min_, key=key)
        elif isinstance(iterable, TupleImpl) and all(_is_num(v) for v in iterable.value):
            if len(iterable.value) == 0:
                if default is not _empty:
                    return default
                raise ValueError("min() arg is an empty sequence")
            return compile_and_call(Array(*iterable.value)._min_, key=key)
        elif isinstance(iterable, SonolusIterator):
            if not (default is _empty or Num._accepts_(default)):
                raise TypeError("default argument must be a number")
            return compile_and_call(
                _min_num_iterator,
                iterable,
                Num._accept_(default) if default is not _empty else None,
                key=key if key is not _identity else None,
            )
        else:
            raise TypeError(f"Unsupported type: {type(iterable)} for min")
    else:
        if default is not _empty:
            raise TypeError("default argument is not supported for min with multiple arguments")
        if not all(_is_num(arg) for arg in args):
            raise TypeError("Arguments to min must be numbers")
        if ctx():
            result = _min2(args[0], args[1], key=key)
            for arg in args[2:]:
                result = _min2(result, arg, key=key)
            return result
        else:
            return min(arg._as_py_() for arg in args)


def _min2(a, b, key=_identity):
    from sonolus.backend.visitor import compile_and_call

    a = validate_value(a)
    b = validate_value(b)
    if _is_num(a) and _is_num(b) and key == _identity:
        return compile_and_call(_min2_num, a, b)
    return compile_and_call(_min2_generic, a, b, key=key)


@native_function(Op.Min)
def _min2_num(a, b):
    if a < b:
        return a
    else:
        return b


def _min2_generic(a, b, key=_identity):
    if key(a) < key(b):
        return a
    else:
        return b


def _min_num_iterator(iterable, default, key):
    iterator = iterable.__iter__()  # noqa: PLC2801
    initial = iterator.next()
    if initial.is_nothing:
        require(default is not None, "default must be provided if the iterator is empty")
        return default
    if key is not None:
        result = initial.get_unsafe()
        best_key = key(result)
        for value in iterator:
            new_key = key(value)
            if new_key < best_key:
                result = value
                best_key = new_key
        return result
    else:
        result = initial.get_unsafe()
        for value in iterator:
            if value < result:  # noqa: PLR1730
                result = value
        return result


@meta_fn
def _callable(value):
    return callable(value)


def _map(fn, iterable, *iterables):
    if len(iterables) == 0:
        return _MappingIterator(fn, iterable.__iter__())  # noqa: PLC2801
    return _MappingIterator(lambda args: fn(*args), zip(iterable, *iterables))  # noqa: B905


def _filter(fn, iterable):
    if fn is None:
        fn = _identity
    return _FilteringIterator(fn, iterable.__iter__())  # noqa: PLC2801


@meta_fn
def _int(value=0):
    value = validate_value(value)
    if not _is_num(value):
        raise TypeError("Only numeric arguments to int() are supported")
    return _trunc(value)


@meta_fn
def _float(value=0.0):
    value = validate_value(value)
    if not _is_num(value):
        raise TypeError("Only numeric arguments to float() are supported")
    return value


def _bool(value=False):
    # Relies on the compiler to perform the conversion in a boolean context
    if value:  # noqa: SIM103
        return True
    else:
        return False


_int._type_mapping_ = Num  # type: ignore
_float._type_mapping_ = Num  # type: ignore
_bool._type_mapping_ = Num  # type: ignore


def _any(iterable):
    for value in iterable:  # noqa: SIM110
        if value:
            return True
    return False


def _all(iterable):
    for value in iterable:  # noqa: SIM110
        if not value:
            return False
    return True


def _sum(iterable, /, start=0):
    for value in iterable:
        start += value
    return start


def _next(iterator):
    require(isinstance(iterator, SonolusIterator), "Only subclasses of SonolusIterator are supported as iterators")
    value = iterator.next()
    if value.is_some:
        return value.get_unsafe()
    error("Iterator has been exhausted")


def _iter(iterable):
    return iterable.__iter__()  # type: ignore # noqa: PLC2801


@meta_fn
def _super(*args):
    """Get the super class of a class or instance."""
    return super(*(arg._as_py_() if arg._is_py_() else arg for arg in args))


@meta_fn
def _hasattr(obj: Any, name: str) -> bool:
    from sonolus.script.internal.constant import ConstantValue
    from sonolus.script.internal.descriptor import SonolusDescriptor

    name = validate_value(name)._as_py_()
    if isinstance(obj, ConstantValue):
        # Unwrap so we can access fields
        obj = obj._as_py_()
    descriptor = None
    for cls in type.mro(type(obj)):
        descriptor = cls.__dict__.get(name, None)
        if descriptor is not None:
            break
    # We want to mirror what getattr supports and fail fast if a future getattr would fail.
    match descriptor:
        case None:
            return hasattr(obj, name)
        case property() | SonolusDescriptor() | FunctionType() | classmethod() | staticmethod():
            return True
        case non_descriptor if not hasattr(non_descriptor, "__get__"):
            return True
        case _:
            raise TypeError(f"Unsupported field or descriptor {name}")


@meta_fn
def _getattr(obj: Any, name: str, default=_empty) -> Any:
    from sonolus.backend.visitor import compile_and_call
    from sonolus.script.internal.constant import ConstantValue
    from sonolus.script.internal.descriptor import SonolusDescriptor

    name = validate_value(name)._as_py_()
    if isinstance(obj, ConstantValue):
        # Unwrap so we can access fields
        obj = obj._as_py_()
    descriptor = None
    for cls in type.mro(type(obj)):
        descriptor = cls.__dict__.get(name, None)
        if descriptor is not None:
            break
    match descriptor:
        case property(fget=getter):
            return compile_and_call(getter, obj)
        case SonolusDescriptor() | FunctionType() | classmethod() | staticmethod() | None:
            return validate_value(getattr(obj, name) if default is _empty else getattr(obj, name, default))
        case non_descriptor if not hasattr(non_descriptor, "__get__"):
            return validate_value(getattr(obj, name) if default is _empty else getattr(obj, name, default))
        case _:
            raise TypeError(f"Unsupported field or descriptor {name}")


@meta_fn
def _setattr(obj: Any, name: str, value: Any):
    from sonolus.backend.visitor import compile_and_call
    from sonolus.script.internal.descriptor import SonolusDescriptor

    name = validate_value(name)._as_py_()
    if obj._is_py_():
        obj = obj._as_py_()
    descriptor = getattr(type(obj), name, None)
    match descriptor:
        case property(fset=setter):
            if setter is None:
                raise AttributeError(f"Cannot set attribute {name} because property has no setter")
            compile_and_call(setter, obj, value)
        case SonolusDescriptor():
            setattr(obj, name, value)
        case _:
            raise TypeError(f"Unsupported field or descriptor {name}")


class _Type(Record):
    @meta_fn
    def __call__(self, value, /):
        value = validate_value(value)
        if value._is_py_():
            value = value._as_py_()
        return validate_value(type(value))

    def __getitem__(self, item):
        return self


_type = _Type()


@meta_fn
def _assert_never(arg: Never, /):
    error("Expected code to be unreachable")


# classmethod, property, staticmethod are supported as decorators, but not within functions

BUILTIN_IMPLS = {
    id(abs): _abs,
    id(all): _all,
    id(any): _any,
    id(bool): _bool,
    id(callable): _callable,
    id(enumerate): _enumerate,
    id(filter): _filter,
    id(float): _float,
    id(getattr): _getattr,
    id(hasattr): _hasattr,
    id(int): _int,
    id(isinstance): _isinstance,
    id(iter): _iter,
    id(len): _len,
    id(map): _map,
    id(max): _max,
    id(min): _min,
    id(next): _next,
    id(range): Range,
    id(reversed): _reversed,
    id(setattr): _setattr,
    id(sum): _sum,
    id(super): _super,
    id(type): _type,
    id(zip): _zip,
    id(assert_never): _assert_never,
    **MATH_BUILTIN_IMPLS,  # Includes round
    **RANDOM_BUILTIN_IMPLS,
}



from __future__ import annotations

import gzip
import hashlib
import json
import urllib.request
import warnings
import zipfile
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Any, Literal, TypedDict, TypeGuard

type Category = Literal[
    "posts",
    "playlists",
    "levels",
    "replays",
    "skins",
    "backgrounds",
    "effects",
    "particles",
    "engines",
]
type Asset = bytes | PathLike | str
CATEGORY_NAMES = {"posts", "playlists", "levels", "replays", "skins", "backgrounds", "effects", "particles", "engines"}
SINGULAR_CATEGORY_NAMES: dict[Category, str] = {
    "posts": "post",
    "playlists": "playlist",
    "levels": "level",
    "replays": "replay",
    "skins": "skin",
    "backgrounds": "background",
    "effects": "effect",
    "particles": "particle",
    "engines": "engine",
}
BASE_PATH = "/sonolus/"
RESERVED_FILENAMES = {"info", "list"}
LOCALIZED_KEYS = {"title", "subtitle", "author", "description", "artists"}
CATEGORY_SORT_ORDER = {
    "levels": 0,
    "engines": 1,
    "skins": 2,
    "effects": 3,
    "particles": 4,
    "backgrounds": 5,
    "posts": 6,
    "playlists": 7,
    "replays": 8,
}


class Collection:
    def __init__(self) -> None:
        self.name = "Unnamed"
        self.categories: dict[Category, dict[str, Any]] = {}
        self.repository: dict[str, bytes] = {}

    def get_item(self, category: Category, name: str) -> Any:
        if name not in self.categories.get(category, {}):
            raise KeyError(f"Item '{name}' not found in category '{category}'")
        return self.categories[category][name]["item"]

    def get_default_item(self, category: Category) -> Any:
        if not self.categories.get(category):
            raise KeyError(f"No items found in category '{category}'")
        return next(iter(self.categories[category].values()))["item"]

    def add_item(self, category: Category, name: str, item: Any) -> None:
        self.categories.setdefault(category, {})[name] = self._make_item_details(item)

    @classmethod
    def _make_item_details(cls, item: dict[str, Any]) -> dict[str, Any]:
        return {
            "item": cls._localize_item(item),
            "actions": [],
            "hasCommunity": False,
            "leaderboards": [],
            "sections": [],
        }

    @staticmethod
    def _load_data(value: Asset) -> bytes:
        return load_asset(value)

    def add_asset(self, value: Asset, /) -> Srl:
        data = self._load_data(value)
        key = hashlib.sha1(data).hexdigest()
        self.repository[key] = data
        return Srl(hash=key, url=f"{BASE_PATH}repository/{key}")

    def load_from_scp(self, zip_data: Asset) -> None:
        with zipfile.ZipFile(BytesIO(self._load_data(zip_data))) as zf:
            files_by_dir = self._group_zip_entries_by_directory(zf.filelist)
            self._process_zip_directories(zf, files_by_dir)

    def load_from_source(self, path: PathLike | str) -> None:
        root_path = Path(path)

        for category_dir in root_path.iterdir():
            if not category_dir.is_dir():
                continue

            category_name = category_dir.name
            if not self._is_valid_category(category_name):
                continue

            for item_dir in category_dir.iterdir():
                if not item_dir.is_dir():
                    continue

                item_json_path = item_dir / "item.json"
                if not item_json_path.exists():
                    continue

                try:
                    item_data = json.loads(item_json_path.read_text(encoding="utf-8"))
                except json.JSONDecodeError:
                    warnings.warn(f"Invalid JSON in {item_json_path}, skipping item.", stacklevel=2)
                    continue

                item_data = self._localize_item(item_data)
                item_data["name"] = item_dir.name

                for resource_path in item_dir.iterdir():
                    if resource_path.name == "item.json":
                        continue

                    try:
                        resource_data = resource_path.read_bytes()

                        if resource_path.suffix.lower() in {".json", ".bin"}:
                            resource_data = gzip.compress(resource_data, mtime=0)

                        srl = self.add_asset(resource_data)
                        item_data[resource_path.stem] = srl

                    except Exception as e:
                        print(f"Error processing resource {resource_path}: {e}")
                        continue

                self.add_item(category_name, item_dir.name, item_data)

    @classmethod
    def _localize_item(cls, item: dict[str, Any]) -> dict[str, Any]:
        localized_item = item.copy()
        for key in LOCALIZED_KEYS:
            if key not in localized_item:
                continue
            localized_item[key] = cls._localize_text(localized_item[key])
        if "tags" in localized_item:
            localized_item["tags"] = [
                {**tag, "title": cls._localize_text(tag["title"])} for tag in localized_item["tags"]
            ]
        localized_item.pop("meta", None)
        return localized_item

    @staticmethod
    def _localize_text(text: str | dict[str, str]) -> str:
        match text:
            case str():
                return text
            case {"en": localized_text}:
                return localized_text
            case {**other_languages} if other_languages:
                return text[min(other_languages)]
            case _:
                return ""

    def _group_zip_entries_by_directory(self, file_list: list[zipfile.ZipInfo]) -> dict[str, list[zipfile.ZipInfo]]:
        files_by_dir: dict[str, list[zipfile.ZipInfo]] = {}

        for zip_entry in file_list:
            if self._should_skip_zip_entry(zip_entry):
                continue

            path_parts = Path(zip_entry.filename).parts
            if path_parts[0] == "sonolus":
                path_parts = path_parts[1:]

            if not path_parts:
                continue

            dir_name = path_parts[0]
            files_by_dir.setdefault(dir_name, []).append(zip_entry)

        return files_by_dir

    def _should_skip_zip_entry(self, zip_entry: zipfile.ZipInfo) -> bool:
        path = Path(zip_entry.filename)
        if path.parts[0] == "sonolus":
            path = Path(*path.parts[1:])
        return zip_entry.filename.endswith("/") or len(path.parts) < 2 or path.name.lower() in RESERVED_FILENAMES

    def _process_zip_directories(self, zf: zipfile.ZipFile, files_by_dir: dict[str, list[zipfile.ZipInfo]]) -> None:
        for dir_name, zip_entries in files_by_dir.items():
            if dir_name == "repository":
                self._add_repository_items(zf, zip_entries)
            elif self._is_valid_category(dir_name):
                self.categories.setdefault(dir_name, {})
                self._extract_category_items(zf, dir_name, zip_entries)

    def _add_repository_items(self, zf: zipfile.ZipFile, zip_entries: list[zipfile.ZipInfo]) -> None:
        for zip_entry in zip_entries:
            self.repository[Path(zip_entry.filename).name] = zf.read(zip_entry)

    def _is_valid_category(self, category: str) -> TypeGuard[Category]:
        return category in CATEGORY_NAMES

    def _extract_category_items(
        self, zf: zipfile.ZipFile, dir_name: Category, zip_entries: list[zipfile.ZipInfo]
    ) -> None:
        for zip_entry in zip_entries:
            try:
                item_details = json.loads(zf.read(zip_entry).decode("utf-8"))
            except json.JSONDecodeError:
                continue

            path = Path(zip_entry.filename)
            if path.parts[0] == "sonolus":
                path = Path(*path.parts[1:])
            item_name = path.stem

            if self._is_valid_category(dir_name):
                self.categories[dir_name][item_name] = item_details

    def write(self, path: Asset) -> None:
        self.link()
        base_dir = self._create_base_directory(path)
        self._write_main_info(base_dir)
        self._write_category_items(base_dir)
        self._write_repository_items(base_dir)

    def link(self):
        for level_details in self.categories.get("levels", {}).values():
            level = level_details["item"]
            if isinstance(level["engine"], str):
                level["engine"] = self.get_item("engines", level["engine"])
            for key, category in (
                ("useSkin", "skins"),
                ("useBackground", "backgrounds"),
                ("useEffect", "effects"),
                ("useParticle", "particles"),
            ):
                use_item = level[key]
                if "item" not in use_item:
                    continue
                use_item_value = use_item["item"]
                if isinstance(use_item_value, str) and use_item_value in self.categories.get(category, {}):
                    use_item["item"] = self.get_item(category, use_item_value)
                elif isinstance(use_item_value, dict) and "name" in use_item_value:
                    name = use_item["item"]["name"]
                    if name in self.categories.get(category, {}):
                        use_item["item"] = self.get_item(category, name)

    def _create_base_directory(self, path: Asset) -> Path:
        base_dir = Path(path) / BASE_PATH.strip("/")
        base_dir.mkdir(parents=True, exist_ok=True)
        return base_dir

    def _write_main_info(self, base_dir: Path) -> None:
        sorted_categories = sorted(self.categories.keys(), key=lambda c: CATEGORY_SORT_ORDER.get(c, 100))
        info = {
            "title": self.name,
            "buttons": [{"type": SINGULAR_CATEGORY_NAMES[category]} for category in sorted_categories],
            "configuration": {"options": []},
        }
        self._write_json(base_dir / "info", info)

    def _write_category_items(self, base_dir: Path) -> None:
        for category, items in self.categories.items():
            if not items:
                continue
            category_dir = self._create_category_directory(base_dir, category)
            self._write_category_structure(category_dir, category, items)

    def _create_category_directory(self, base_dir: Path, category: Category) -> Path:
        category_dir = base_dir / category
        category_dir.mkdir(exist_ok=True)
        return category_dir

    def _write_category_structure(self, category_dir: Path, category: Category, items: dict[str, Any]) -> None:
        self._write_json(
            category_dir / "info",
            {
                "sections": [
                    {
                        "itemType": SINGULAR_CATEGORY_NAMES[category],
                        "title": "Items",
                        "items": [item_details["item"] for item_details in items.values()],
                    }
                ]
            },
        )

        category_list = {"pageCount": 1, "items": [item_details["item"] for item_details in items.values()]}
        self._write_json(category_dir / "list", category_list)

        for item_name, item_details in items.items():
            self._write_json(category_dir / item_name, item_details)

    def _write_repository_items(self, base_dir: Path) -> None:
        repo_dir = base_dir / "repository"
        repo_dir.mkdir(exist_ok=True)

        for key, data in self.repository.items():
            (repo_dir / key).write_bytes(data)

    @staticmethod
    def _write_json(path: Path, content: Any) -> None:
        path.write_text(json.dumps(content), encoding="utf-8")

    def update(self, other: Collection) -> None:
        self.repository.update(other.repository)
        for category, items in other.categories.items():
            self.categories.setdefault(category, {}).update(items)


class Srl(TypedDict):
    hash: str
    url: str


def load_asset(value: Asset) -> bytes:
    match value:
        case str() if value.startswith(("http://", "https://")):
            headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.9",
                "Connection": "keep-alive",
            }
            request = urllib.request.Request(value, headers=headers)
            with urllib.request.urlopen(request) as response:
                return response.read()
        case PathLike():
            return Path(value).read_bytes()
        case bytes():
            return value
        case _:
            raise TypeError("value must be a URL, a path, or bytes")



from sonolus.script.debug import debug_log
from tests.script.conftest import run_and_validate


class A:
    def m(self):
        debug_log(1)

    @classmethod
    def m_cls(cls):
        debug_log(1)


class B(A):
    def m(self):
        debug_log(2)
        super().m()
        debug_log(3)

    @classmethod
    def m_cls(cls):
        debug_log(2)
        super().m_cls()
        debug_log(3)


class C(A):
    def m(self):
        debug_log(4)
        super().m()
        debug_log(5)

    @classmethod
    def m_cls(cls):
        debug_log(4)
        super().m_cls()
        debug_log(5)


class D(B, C):
    def m(self):
        debug_log(6)
        super().m()
        debug_log(7)

    @classmethod
    def m_cls(cls):
        debug_log(6)
        super().m_cls()
        debug_log(7)


def test_super_simple():
    b = B()
    b._is_comptime_value_ = True  # type: ignore

    def fn():
        b.m()

    run_and_validate(fn)


def test_super_diamond():
    d = D()
    d._is_comptime_value_ = True  # type: ignore

    def fn():
        d.m()

    run_and_validate(fn)


def test_super_classmethod():
    b = B()
    b._is_comptime_value_ = True  # type: ignore

    def fn():
        b.m_cls()

    run_and_validate(fn)


def test_super_classmethod_diamond():
    d = D()
    d._is_comptime_value_ = True  # type: ignore

    def fn():
        d.m_cls()

    run_and_validate(fn)



# ruff: noqa
from typing import (
    Any,
    Sequence,
    overload,
    MutableSequence,
)

@overload
def randrange(stop: int) -> int:
    """Return a randomly selected element from range(stop).

    Args:
        stop: The end of the range.

    Returns:
        A randomly selected integer from the range.
    """
    ...

@overload
def randrange(start: int, stop: int, step: int = ...) -> int:
    """Return a randomly selected element from range(start, stop, step).

    Args:
        start: The start of the range.
        stop: The end of the range.
        step: The step size.

    Returns:
        A randomly selected integer from the range.
    """
    ...

def randrange(start: int, stop: int = ..., step: int = ...) -> int:
    """Return a randomly selected element from range(start, stop, step).

    Args:
        start: The start of the range.
        stop: The end of the range.
        step: The step size.

    Returns:
        A randomly selected integer from the range.
    """
    ...

def randint(a: int, b: int) -> int:
    """Return a random integer N such that a <= N <= b.

    Args:
        a: The lower bound.
        b: The upper bound.

    Returns:
        A randomly selected integer between a and b, inclusive.
    """
    ...

def choice[T](seq: Sequence[T]) -> T:
    """Return a randomly selected element from a non-empty sequence.

    Args:
        seq: The sequence to choose from.

    Returns:
        A randomly selected element from the sequence.
    """
    ...

def shuffle(seq: MutableSequence[Any]) -> None:
    """Shuffle the sequence in place.

    Args:
        seq: The mutable sequence to shuffle.
    """
    ...

def random() -> float:
    """Return a random floating point number in the range [0.0, 1.0).

    Returns:
        A random float between 0.0 (inclusive) and 1.0 (exclusive).
    """
    ...

def uniform(a: float, b: float) -> float:
    """Return a random floating point number N such that a <= N <= b.

    Args:
        a: The lower bound.
        b: The upper bound.

    Returns:
        A random float between a and b.
    """
    ...



from math import isfinite, isinf, isnan

from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.node import EngineNode, FunctionNode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_reverse_postorder
from sonolus.backend.place import BlockPlace


def cfg_to_engine_node(entry: BasicBlock):
    block_indexes = {block: i for i, block in enumerate(traverse_cfg_reverse_postorder(entry))}
    block_statements = []
    for block in block_indexes:
        statements = []
        statements.extend(ir_to_engine_node(stmt) for stmt in block.statements)
        outgoing = {
            edge.cond: edge.dst for edge in sorted(block.outgoing, key=lambda edge: (edge.cond is None, edge.cond))
        }
        match outgoing:
            case {**other} if not other:
                statements.append(len(block_indexes))
            case {None: target, **other} if not other:
                statements.append(block_indexes[target])
            case {0: f_branch, None: t_branch, **other} if not other:
                statements.append(
                    FunctionNode(
                        func=Op.If,
                        args=(
                            ir_to_engine_node(block.test),
                            block_indexes[t_branch],
                            block_indexes[f_branch],
                        ),
                    )
                )
            case {None: default_branch, **other} if len(other) == 1:
                cond, cond_branch = next(iter(other.items()))
                statements.append(
                    FunctionNode(
                        func=Op.If,
                        args=(
                            ir_to_engine_node(IRPureInstr(Op.Equal, args=[block.test, IRConst(cond)])),
                            block_indexes[cond_branch],
                            block_indexes[default_branch],
                        ),
                    )
                )
            case dict() as targets:
                args = [ir_to_engine_node(block.test)]
                default = len(block_indexes)
                conds = [cond for cond in targets if cond is not None]
                if min(conds) == 0 and max(conds) == len(conds) - 1 and all(int(cond) == cond for cond in conds):
                    args.extend(block_indexes[targets[cond]] for cond in range(len(conds)))
                    if None in targets:
                        default = block_indexes[targets[None]]
                    args.append(default)
                    statements.append(FunctionNode(Op.SwitchIntegerWithDefault, tuple(args)))
                else:
                    for cond, target in targets.items():
                        if cond is None:
                            default = block_indexes[target]
                            continue
                        args.append(cond)
                        args.append(block_indexes[target])
                    args.append(default)
                    statements.append(FunctionNode(Op.SwitchWithDefault, tuple(args)))
        block_statements.append(FunctionNode(Op.Execute, tuple(statements)))
    block_statements.append(0)
    result = FunctionNode(Op.Block, (FunctionNode(Op.JumpLoop, tuple(block_statements)),))
    for block in block_indexes:
        # Clean up without relying on gc
        del block.incoming
        del block.outgoing
        del block.phis
        del block.statements
        del block.test
    return result


def ir_to_engine_node(stmt) -> EngineNode:
    match stmt:
        case int(value) | float(value) | IRConst(value=int(value) | float(value)):
            value = float(value)
            if value.is_integer():
                return int(value)
            elif isfinite(value):
                return value
            elif isinf(value):
                # Read values from ROM
                return FunctionNode(Op.Get, args=(3000, 1 if value > 0 else 2))
            elif isnan(value):
                # Read value from ROM
                return FunctionNode(Op.Get, args=(3000, 0))
            else:
                raise ValueError(f"Invalid constant value: {value}")
        case IRPureInstr(op=op, args=args) | IRInstr(op=op, args=args):
            return FunctionNode(func=op, args=tuple(ir_to_engine_node(arg) for arg in args))
        case IRGet(place=place):
            return ir_to_engine_node(place)
        case BlockPlace() as place:
            if place.offset == 0:
                index = ir_to_engine_node(place.index)
            elif place.index == 0:
                index = place.offset
            else:
                index = FunctionNode(func=Op.Add, args=(ir_to_engine_node(place.index), place.offset))
            return FunctionNode(func=Op.Get, args=(ir_to_engine_node(place.block), index))
        case IRSet(place=place, value=value):
            return FunctionNode(func=Op.Set, args=(*ir_to_engine_node(place).args, ir_to_engine_node(value)))
        case _:
            raise TypeError(f"Unsupported IR statement: {stmt}")



from sonolus.backend.blocks import BlockData
from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet, IRStmt
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock

RUNTIME_CONSTANT_BLOCKS = {
    "RuntimeEnvironment",
    "RuntimeUI",
    "RuntimeUIConfiguration",
    "LevelData",
    "LevelOption",
    "LevelBucket",
    "LevelScore",
    "LevelLife",
    "EngineRom",
    "ArchetypeLife",
    "RuntimeCanvas",
    "PreviewData",
    "PreviewOption",
    "TutorialData",
}


class InlineVars(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        use_counts: dict[SSAPlace, int] = {}
        definitions: dict[SSAPlace, IRStmt] = {}

        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                self.count_uses(stmt, use_counts)
                if isinstance(stmt, IRSet) and isinstance(stmt.place, SSAPlace):
                    definitions[stmt.place] = stmt.value
            self.count_uses(block.test, use_counts)
            for tgt, args in block.phis.items():
                for arg in args.values():
                    self.count_uses(arg, use_counts)
                if len(args) == 1:
                    arg = next(iter(args.values()))
                    definitions[tgt] = IRGet(place=arg)

        for defn in definitions.values():
            if isinstance(defn, IRGet) and isinstance(defn.place, SSAPlace):
                use_counts[defn.place] -= 1

        canonical_definitions: dict[SSAPlace, IRStmt] = {}
        for p, defn in definitions.items():
            canonical_definitions[p] = defn
            # Update the definition if it's a Get from another SSAPlace until we reach a definition that is not a Get
            while defn and isinstance(defn, IRGet) and isinstance(defn.place, SSAPlace):
                canonical_definitions[p] = defn
                defn = definitions.get(defn.place)  # Can be None if it's a phi
            canonical_defn = canonical_definitions[p]
            if (
                use_counts.get(p, 0) > 0
                and isinstance(canonical_defn, IRGet)
                and isinstance(canonical_defn.place, SSAPlace)
            ):
                use_counts[canonical_defn.place] = use_counts.get(canonical_defn.place, 0) + 1

        for p, defn in canonical_definitions.items():
            if isinstance(defn, IRGet) and isinstance(defn.place, SSAPlace):
                inner_p = defn.place
                inner_defn = canonical_definitions.get(inner_p)
                if (
                    inner_defn
                    and self.is_inlinable(inner_defn, config.callback)
                    and (use_counts.get(inner_p, 0) <= 1 or self.is_free_to_inline(inner_defn, config.callback))
                ):
                    canonical_definitions[p] = inner_defn

        inlined_definitions = {**canonical_definitions}
        for p, defn in canonical_definitions.items():
            while True:
                inlinable_uses = self.get_inlinable_uses(defn, set())
                subs = {}
                for inside_p in inlinable_uses:
                    if inside_p not in canonical_definitions:
                        continue
                    inside_defn = canonical_definitions[inside_p]
                    if not self.is_inlinable(inside_defn, config.callback):
                        continue
                    if (
                        (isinstance(inside_defn, IRGet) and isinstance(inside_defn.place, SSAPlace))
                        or use_counts[inside_p] == 1
                        or self.is_free_to_inline(inside_defn, config.callback)
                    ):
                        subs[inside_p] = inside_defn
                if not subs:
                    break
                defn = self.substitute(defn, subs)
            inlined_definitions[p] = defn

        valid = {
            p
            for p in inlined_definitions
            if self.is_inlinable(inlined_definitions[p], config.callback)
            and (use_counts.get(p, 0) <= 1 or self.is_free_to_inline(inlined_definitions[p], config.callback))
        }

        for block in traverse_cfg_preorder(entry):
            new_statements = []
            for stmt in [*block.statements, block.test]:
                if (
                    isinstance(stmt, IRSet)
                    and isinstance(stmt.place, SSAPlace)
                    and isinstance(stmt.value, IRGet)
                    and isinstance(stmt.value.place, SSAPlace)
                ):
                    # Don't bother inlining a direct alias since it can get optimized away later and
                    # reordering can reduce optimality since we don't have many other code motion optimizations.
                    new_statements.append(stmt)
                    continue
                while True:
                    inlinable_uses = self.get_inlinable_uses(stmt, set())
                    subs = {}
                    for p in inlinable_uses:
                        if p in valid:
                            subs[p] = inlined_definitions[p]

                    if subs:
                        stmt = self.substitute(stmt, subs)
                    else:
                        new_statements.append(stmt)
                        break

            block.statements = new_statements[:-1]
            block.test = new_statements[-1]

        return entry

    def substitute(self, stmt, subs):
        match stmt:
            case IRConst():
                return stmt
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.substitute(arg, subs) for arg in args])
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.substitute(arg, subs) for arg in args])
            case IRGet(place=place):
                if place in subs:
                    return subs[place]
                return stmt
            case IRSet(place=place, value=value):
                return IRSet(place=place, value=self.substitute(value, subs))
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def count_uses(self, stmt, use_counts):
        match stmt:
            case IRConst():
                pass
            case IRInstr(op=_, args=args) | IRPureInstr(op=_, args=args):
                for arg in args:
                    self.count_uses(arg, use_counts)
            case IRGet(place=place):
                self.count_uses(place, use_counts)
            case IRSet(place=place, value=value):
                if not isinstance(place, SSAPlace):  # We don't want to count the definition itself
                    self.count_uses(place, use_counts)
                self.count_uses(value, use_counts)
            case SSAPlace():
                use_counts[stmt] = use_counts.get(stmt, 0) + 1
            case BlockPlace(block=block, index=index, offset=_):
                self.count_uses(block, use_counts)
                self.count_uses(index, use_counts)
            case int() | float():
                pass
            case TempBlock():
                pass
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")
        return use_counts

    def get_inlinable_uses(self, stmt, uses):
        match stmt:
            case IRConst():
                pass
            case IRInstr(op=_, args=args) | IRPureInstr(op=_, args=args):
                for arg in args:
                    self.get_inlinable_uses(arg, uses)
            case IRGet(place=place):
                if isinstance(place, SSAPlace):
                    uses.add(place)
            case IRSet(place=_, value=value):
                self.get_inlinable_uses(value, uses)
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")
        return uses

    def is_inlinable(self, stmt, callback: str):
        match stmt:
            case IRConst():
                return True
            case IRInstr(op=op, args=args) | IRPureInstr(op=op, args=args):
                return not op.side_effects and op.pure and all(self.is_inlinable(arg, callback) for arg in args)
            case IRGet():
                return isinstance(stmt.place, SSAPlace) or (
                    isinstance(stmt.place, BlockPlace)
                    and isinstance(stmt.place.block, BlockData)
                    and callback not in stmt.place.block.writable
                    and isinstance(stmt.place.index, int | SSAPlace)
                )
            case IRSet():
                return False
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def is_free_to_inline(self, stmt: IRStmt, callback: str) -> bool:
        match stmt:
            case IRConst():
                return True
            case IRInstr() | IRPureInstr():
                return self.is_runtime_constant(stmt, callback)
            case IRGet():
                return isinstance(stmt.place, SSAPlace) or (
                    isinstance(stmt.place, BlockPlace)
                    and isinstance(stmt.place.block, float | int)
                    and isinstance(stmt.place.index, float | int)
                )
            case IRSet():
                return False
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def is_runtime_constant(self, stmt: IRStmt, callback: str) -> bool:
        match stmt:
            case IRConst():
                return True
            case IRInstr(op=op, args=args) | IRPureInstr(op=op, args=args):
                return not op.side_effects and op.pure and all(self.is_runtime_constant(arg, callback) for arg in args)
            case IRGet():
                return (
                    isinstance(stmt.place, BlockPlace)
                    and isinstance(stmt.place.block, BlockData)
                    and callback not in stmt.place.block.writable
                    and stmt.place.block.name in RUNTIME_CONSTANT_BLOCKS
                    and isinstance(stmt.place.index, int | SSAPlace)
                )
            case IRSet():
                return False
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")



from pydori.lib.layer import LAYER_CONNECTOR, LAYER_SIM_LINE, get_z
from pydori.lib.layout import layout_hold_connector, layout_sim_line, note_y_to_alpha
from pydori.lib.options import Options
from pydori.lib.skin import Skin


def draw_hold_connector(
    lane_a: float,
    lane_b: float,
    y_a: float,
    y_b: float,
):
    layout = layout_hold_connector(lane_a, lane_b, y_a, y_b)
    sprite = Skin.hold_connector
    sprite.draw(layout, z=get_z(LAYER_CONNECTOR, lane=min(lane_a, lane_b), y=min(y_a, y_b)), a=Options.connector_alpha)


def draw_sim_line(
    lane_a: float,
    lane_b: float,
    y: float,
):
    if not Options.sim_lines_enabled:
        return
    alpha = Options.sim_line_alpha * note_y_to_alpha(y)
    if alpha <= 0:
        return
    layout = layout_sim_line(lane_a, lane_b, y)
    sprite = Skin.sim_line
    sprite.draw(layout, z=get_z(LAYER_SIM_LINE, lane=min(lane_a, lane_b), y=y), a=alpha)



from __future__ import annotations

from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import _merge_sort_linked_list_nodes  # noqa: PLC2701
from sonolus.script.record import Record
from tests.script.conftest import run_and_validate


class Node[T](Record):
    value: T
    next_index: int


class NodeRef[T, Size](Record):
    arr: Array[Node[T], Size]
    index: int

    def get_value(self) -> T:
        return self.arr[self.index].value

    def get_next(self) -> NodeRef:
        next_index = self.arr[self.index].next_index
        return NodeRef(self.arr, next_index)

    def set_next(self, next_node: NodeRef) -> None:
        self.arr[self.index].next_index = next_node.index

    def set_prev(self, prev_node: NodeRef) -> None:
        pass  # Not needed

    def is_present(self) -> bool:
        return self.index != -1

    def set(self, other: NodeRef) -> None:
        self.index = other.index

    def copy(self) -> NodeRef:
        return NodeRef(self.arr, self.index)

    def empty(self) -> NodeRef:
        return NodeRef(self.arr, -1)


@given(
    elts=st.lists(st.integers(min_value=-9999, max_value=9999), min_size=0, max_size=100),
)
def test_merge_sort_linked_list_nodes(elts: list[int]):
    orig_values = Array[Node[int], len(elts)](
        *(Node(value=elt, next_index=i + 1 if i + 1 < len(elts) else -1) for i, elt in enumerate(elts))
    )

    size = len(elts)

    def fn():
        values = +orig_values
        head_ref = NodeRef(values, 0 if size > 0 else -1)

        sorted_head_ref = _merge_sort_linked_list_nodes(head_ref)

        results = +Array[int, size]
        current_ref = sorted_head_ref.copy()
        idx = 0
        while current_ref.is_present():
            results[idx] = current_ref.get_value()
            current_ref.set(current_ref.get_next())
            idx += 1
        return results

    assert list(run_and_validate(fn)) == sorted(elts)



from __future__ import annotations

import gzip
import json
from collections.abc import Iterator
from os import PathLike
from pathlib import Path
from typing import Any, NamedTuple, NotRequired, TypedDict

from sonolus.build.collection import Asset, load_asset
from sonolus.script.archetype import PlayArchetype, StandardArchetypeName, StandardImport
from sonolus.script.metadata import AnyText, Tag, as_localization_text


class ExportedLevel:
    """An exported Sonolus level."""

    def __init__(
        self,
        *,
        item: dict,
        cover: bytes,
        bgm: bytes,
        preview: bytes | None,
        data: bytes,
    ):
        self.item = item
        self.cover = cover
        self.bgm = bgm
        self.preview = preview
        self.data = data

    def write_to_dir(self, path: PathLike):
        """Write the exported level to a directory."""
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        (path / "item.json").write_text(json.dumps(self.item, ensure_ascii=False), encoding="utf-8")
        (path / "cover").write_bytes(self.cover)
        (path / "bgm").write_bytes(self.bgm)
        if self.preview is not None:
            (path / "preview").write_bytes(self.preview)
        (path / "data").write_bytes(self.data)


class Level:
    """A Sonolus level.

    Args:
        name: The name of the level.
        title: The title of the level.
        rating: The rating of the level.
        artists: The artists of the level.
        author: The author of the level.
        cover: The cover of the level.
        bgm: The background music of the level.
        data: The data of the level.
        use_skin: The skin to use, overriding the engine skin.
        use_background: The background to use, overriding the engine background.
        use_effect: The effect to use, overriding the engine effect.
        use_particle: The particle to use, overriding the engine particle.
        tags: The tags of the level.
        description: The description of the level.
        meta: Additional metadata of the level.
    """

    version = 1

    def __init__(
        self,
        *,
        name: str,
        title: AnyText | None = None,
        rating: int = 0,
        artists: AnyText = "Unknown",
        author: AnyText = "Unknown",
        cover: Asset | None = None,
        bgm: Asset | None = None,
        preview: Asset | None = None,
        data: LevelData,
        use_skin: str | None = None,
        use_background: str | None = None,
        use_effect: str | None = None,
        use_particle: str | None = None,
        tags: list[Tag] | None = None,
        description: AnyText | None = None,
        meta: Any = None,
    ) -> None:
        self.name = name
        self.title = as_localization_text(title or name)
        self.rating = rating
        self.artists = as_localization_text(artists)
        self.author = as_localization_text(author)
        self.cover = cover
        self.bgm = bgm
        self.preview = preview
        self.data = data
        self.use_skin = use_skin
        self.use_background = use_background
        self.use_effect = use_effect
        self.use_particle = use_particle
        self.tags = tags or []
        self.description = as_localization_text(description) if description is not None else None
        self.meta = meta

    def export(self, engine_name: str) -> ExportedLevel:
        """Export the level in a sonolus-pack compatible format.

        Args:
            engine_name: The name of the engine this level is for.

        Returns:
            The exported level.
        """
        from sonolus.build.level import package_level_data
        from sonolus.build.project import BLANK_AUDIO, BLANK_PNG

        item = {
            "version": self.version,
            "rating": self.rating,
            "engine": engine_name,
            "useSkin": {"useDefault": True} if self.use_skin is None else {"useDefault": False, "item": self.use_skin},
            "useBackground": {"useDefault": True}
            if self.use_background is None
            else {"useDefault": False, "item": self.use_background},
            "useEffect": {"useDefault": True}
            if self.use_effect is None
            else {"useDefault": False, "item": self.use_effect},
            "useParticle": {"useDefault": True}
            if self.use_particle is None
            else {"useDefault": False, "item": self.use_particle},
            "title": self.title,
            "artists": self.artists,
            "author": self.author,
            "tags": [tag.as_dict() for tag in self.tags],
        }
        if self.description is not None:
            item["description"] = self.description
        if self.meta is not None:
            item["meta"] = self.meta
        return ExportedLevel(
            item=item,
            cover=load_asset(self.cover) if self.cover is not None else BLANK_PNG,
            bgm=load_asset(self.bgm) if self.bgm is not None else BLANK_AUDIO,
            preview=load_asset(self.preview) if self.preview is not None else None,
            data=package_level_data(self.data),
        )


type EntityListArg = list[list[PlayArchetype] | PlayArchetype] | PlayArchetype


def flatten_entities(entities: EntityListArg) -> Iterator[PlayArchetype]:
    """Flatten a list of entities.

    Args:
        entities: The list of entities.

    Yields:
        The flattened entities.
    """
    if isinstance(entities, list):
        for entity in entities:
            yield from flatten_entities(entity)
    else:
        yield entities


class LevelData:
    """The data of a Sonolus level.

    Args:
        bgm_offset: The background music audio offset.
        entities: The entities of the level.
    """

    bgm_offset: float
    entities: list[PlayArchetype]

    def __init__(self, bgm_offset: float, entities: EntityListArg) -> None:
        self.bgm_offset = bgm_offset
        self.entities = [*flatten_entities(entities)]


class BpmChange(PlayArchetype):
    """The standard bpm change archetype."""

    name = StandardArchetypeName.BPM_CHANGE

    beat: StandardImport.BEAT
    bpm: StandardImport.BPM


class TimescaleChange(PlayArchetype):
    """The standard timescale change archetype."""

    name = StandardArchetypeName.TIMESCALE_CHANGE

    beat: StandardImport.BEAT
    timescale: StandardImport.TIMESCALE


class ExternalLevelDataDict(TypedDict):
    bgmOffset: float
    entities: list[ExternalEntityDataDict]


class ExternalEntityDataDict(TypedDict):
    name: NotRequired[str]
    archetype: str
    data: NotRequired[list[ExternalEntityDataValueDict]]


class ExternalEntityDataValueDict(TypedDict):
    name: str
    value: NotRequired[int | float]
    ref: NotRequired[str]


class ExternalLevelData(NamedTuple):
    """Level data parsed from an external source."""

    bgm_offset: float
    entities: list[ExternalEntityData]


class ExternalEntityData(NamedTuple):
    """Entity data parsed from an external source."""

    archetype: str
    data: dict[str, Any]


def parse_external_level_data(raw_data: ExternalLevelDataDict | str | bytes, /) -> ExternalLevelData:
    """Parse level data from an external source.

    If given a string, it is parsed as JSON. If given bytes, it is un-gzipped and then parsed as JSON.

    Args:
        raw_data: The raw level data to parse.

    Returns:
        The parsed level data.
    """
    if isinstance(raw_data, bytes):
        raw_data = gzip.decompress(raw_data).decode("utf-8")
    if isinstance(raw_data, str):
        raw_data = json.loads(raw_data)
    bgm_offset = raw_data["bgmOffset"]
    raw_entities = raw_data["entities"]
    entity_name_to_index = {e["name"]: i for i, e in enumerate(raw_entities) if "name" in e}
    entities = []
    for raw_entity in raw_entities:
        archetype = raw_entity["archetype"]
        data = {}
        for entry in raw_entity.get("data", []):
            if "value" in entry:
                data[entry["name"]] = entry["value"]
            elif "ref" in entry:
                data[entry["name"]] = entity_name_to_index.get(entry["ref"], 0)
        entities.append(ExternalEntityData(archetype=archetype, data=data))
    return ExternalLevelData(bgm_offset=bgm_offset, entities=entities)



# ruff: noqa: FURB171
import random

from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.debug import assert_true
from tests.script.conftest import run_and_validate

nums = st.one_of(
    st.integers(min_value=-999, max_value=999),
    st.floats(min_value=-999.0, max_value=999.0, allow_nan=False, allow_infinity=False),
)


@st.composite
def set_and_present_value(draw, fixed_size: int | None = None):
    values = draw(
        st.sets(
            nums, min_size=1 if fixed_size is None else fixed_size, max_size=20 if fixed_size is None else fixed_size
        )
    )
    value = draw(st.sampled_from(list(values)))
    return values, value


@st.composite
def set_and_missing_value(draw, fixed_size: int | None = None):
    values = draw(
        st.sets(
            nums, min_size=1 if fixed_size is None else fixed_size, max_size=20 if fixed_size is None else fixed_size
        )
    )
    missing = draw(nums.filter(lambda x: x not in values))
    return values, missing


def test_contains_basic():
    def fn():
        assert_true(2 in {1, 2, 3, 4, 5})
        assert_true(5 in {1, 2, 3, 4, 5})
        assert_true(6 not in {1, 2, 3, 4, 5})
        assert_true(0 not in {1, 2, 3, 4, 5})
        return 1

    assert run_and_validate(fn) == 1


def test_contains_out_of_order():
    def fn():
        assert_true(3 in {5, 1, 4, 2, 3})
        assert_true(1 in {5, 1, 4, 2, 3})
        assert_true(6 not in {5, 1, 4, 2, 3})
        assert_true(0 not in {5, 1, 4, 2, 3})
        return 1

    assert run_and_validate(fn) == 1


def test_contains_non_literal_contents():
    def fn():
        a = 5 * (random.random() != -1)
        b = 4 * (random.random() != -1)
        c = 3 * (random.random() != -1)
        d = 2 * (random.random() != -1)
        e = 1 * (random.random() != -1)
        test_set = {a, b, c, d, e}
        assert_true(2 in test_set)
        assert_true(5 in test_set)
        assert_true(6 not in test_set)
        assert_true(0 not in test_set)
        return 1

    assert run_and_validate(fn) == 1


def test_defined_outside():
    test_set = {10, 20, 30, 40, 50}

    def fn():
        assert_true(10 in test_set)
        assert_true(30 in test_set)
        assert_true(50 in test_set)
        assert_true(5 not in test_set)
        assert_true(60 not in test_set)
        return 1

    assert run_and_validate(fn) == 1


def test_defined_outside_frozenset():
    test_set = frozenset({10, 20, 30, 40, 50})

    def fn():
        assert_true(10 in test_set)
        assert_true(30 in test_set)
        assert_true(50 in test_set)
        assert_true(5 not in test_set)
        assert_true(60 not in test_set)
        return 1

    assert run_and_validate(fn) == 1


def test_empty():
    empty_set = set()

    def fn():
        assert_true(0 not in empty_set)
        assert_true(1 not in empty_set)
        assert_true(0.0 not in empty_set)
        return 1

    assert run_and_validate(fn) == 1


def test_single_element_set():
    def fn():
        assert_true(42 in {42})
        assert_true(0 not in {42})
        assert_true(41 not in {42})
        assert_true(43 not in {42})
        return 1

    assert run_and_validate(fn) == 1


def test_negative_numbers():
    def fn():
        assert_true(-5 in {-10, -5, 0, 5, 10})
        assert_true(-10 in {-10, -5, 0, 5, 10})
        assert_true(0 in {-10, -5, 0, 5, 10})
        assert_true(-3 not in {-10, -5, 0, 5, 10})
        assert_true(-15 not in {-10, -5, 0, 5, 10})
        return 1

    assert run_and_validate(fn) == 1


@given(set_and_present_value())
def test_set_contains_present(args):
    value_set, value = args

    def fn():
        return value in value_set

    assert run_and_validate(fn)


@given(set_and_missing_value())
def test_set_contains_missing(args):
    value_set, missing = args

    def fn():
        return missing in value_set

    assert not run_and_validate(fn)


@given(set_and_present_value(fixed_size=20))
def test_dyn_set_contains_present(args):
    value_set, value = args

    def fn():
        values = +Array[float, 20]
        for i, v in enumerate(value_set):
            values[i] = v * (random.random() != -1)
        return value in {
            values[0],
            values[1],
            values[2],
            values[3],
            values[4],
            values[5],
            values[6],
            values[7],
            values[8],
            values[9],
            values[10],
            values[11],
            values[12],
            values[13],
            values[14],
            values[15],
            values[16],
            values[17],
            values[18],
            values[19],
        }

    assert run_and_validate(fn)


@given(set_and_missing_value(fixed_size=20))
def test_dyn_set_contains_missing(args):
    value_set, missing = args

    def fn():
        values = +Array[float, 20]
        for i, v in enumerate(value_set):
            values[i] = v * (random.random() != -1)
        return missing in {
            values[0],
            values[1],
            values[2],
            values[3],
            values[4],
            values[5],
            values[6],
            values[7],
            values[8],
            values[9],
            values[10],
            values[11],
            values[12],
            values[13],
            values[14],
            values[15],
            values[16],
            values[17],
            values[18],
            values[19],
        }

    assert not run_and_validate(fn)



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Protocol, Self

from sonolus.backend.visitor import compile_and_call
from sonolus.script.archetype import AnyArchetype, EntityRef
from sonolus.script.array import Array
from sonolus.script.array_like import ArrayLike, get_positive_index
from sonolus.script.debug import error
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn
from sonolus.script.interval import clamp
from sonolus.script.iterator import SonolusIterator
from sonolus.script.maybe import Maybe, Nothing, Some
from sonolus.script.num import Num
from sonolus.script.pointer import _deref
from sonolus.script.record import Record
from sonolus.script.values import copy, zeros


class Box[T](Record):
    """A box that contains a value.

    This can be helpful for generic code that can handle both Num and non-Num types.

    Usage:
        ```python
        Box[T](value: T)
        ```

    Examples:
        ```python
        box = Box(1)
        box = Box[int](2)

        x: T = ...
        y: T = ...
        box = Box(x)
        box.value = y  # Works regardless of whether x is a Num, array, or record
        ```
    """

    value: T
    """The value contained in the box."""


class Pair[T, U](Record):
    """A generic pair of values.

    Usage:
        ```python
        Pair[T, U](first: T, second: U)
        ```

    Examples:
        ```python
        pair = Pair(1, 2)
        pair = Pair[int, Pair[int, int]](1, Pair(2, 3))
        ```
    """

    first: T
    """The first value."""

    second: U
    """The second value."""

    def __lt__(self, other):
        if self.first == other.first:
            return self.second < other.second
        return self.first < other.first

    def __le__(self, other):
        if self.first == other.first:
            return self.second <= other.second
        return self.first <= other.first

    def __gt__(self, other):
        if self.first == other.first:
            return self.second > other.second
        return self.first > other.first

    def __ge__(self, other):
        if self.first == other.first:
            return self.second >= other.second
        return self.first >= other.first


class VarArray[T, Capacity](Record, ArrayLike[T]):
    """An array with a variable size and fixed maximum capacity.

    Supports negative indexes.

    Usage:
        ```python
        VarArray[T, Capacity].new()  # Create a new empty array
        ```

    Examples:
        ```python
        array = VarArray[int, 10].new()
        array.append(1)
        ```
    """

    _size: int
    _array: Array[T, Capacity]

    @classmethod
    def new(cls) -> Self:
        """Create a new empty array."""
        element_type = cls.type_var_value(T)
        capacity = cls.type_var_value(Capacity)
        return cls(0, zeros(Array[element_type, capacity]))

    def __len__(self) -> int:
        """Return the number of elements in the array."""
        return self._size

    @classmethod
    def capacity(cls) -> int:
        """Return the maximum number of elements the array can hold."""
        return cls.type_var_value(Capacity)

    def is_full(self) -> bool:
        """Return whether the array is full."""
        return self._size == self.capacity()

    def __getitem__(self, item) -> T:
        """Return the element at the given index.

        Supports negative indexes.

        The returned value continues to be part of the array.
        Future modifications to the array will affect the returned value.

        Note:
            Future modifications to the array may cause unexpected changes to the returned value.
            If the array may be modified in the future, it's recommended to make a copy of the value.

            For example:
            ```python
            a = VarArray[Pair, 10].new()
            a.append(Pair(1, 2))
            a.append(Pair(3, 4))
            a.append(Pair(5, 6))
            p = a[1]
            a.pop(0)  # Elements are shifted back
            assert p == Pair(5, 6)  # The value of p has changed
            ```
        """
        return self._array.get_unchecked(get_positive_index(item, self._size))

    def __setitem__(self, key: int, value: T):
        """Update the element at the given index."""
        self._array.set_unchecked(get_positive_index(key, self._size), value)

    def __delitem__(self, key: int):
        """Remove the element at the given index."""
        self.pop(key)

    def append(self, value: T):
        """Append a copy of the given value to the end of the array.

        Args:
            value: The value to append.
        """
        assert self._size < len(self._array), "Array is full"
        self._array.set_unchecked(self._size, value)
        self._size += 1

    def append_unchecked(self, value: T):
        """Append the given value to the end of the array without checking the capacity.

        Use with caution as this may cause hard to debug issues if the array is full.

        Args:
            value: The value to append.
        """
        self._array.set_unchecked(self._size, value)
        self._size += 1

    def extend(self, values: ArrayLike[T]):
        """Appends copies of the values in the given array to the end of the array.

        Args:
            values: The values to append.
        """
        assert self._size + len(values) <= len(self._array), "Array is full"
        i = 0
        while i < len(values):
            self._array.set_unchecked(self._size + i, values.get_unchecked(i))
            i += 1
        self._size += len(values)

    def pop(self, index: int | None = None) -> T:
        """Remove and return a copy of the value at the given index.

        Preserves the relative order of the elements.

        Args:
            index: The index of the value to remove. If None, the last element is removed.
        """
        if index is None:
            index = self._size - 1
        index = get_positive_index(index, self._size)
        value = copy(self._array.get_unchecked(index))
        self._size -= 1
        if index < self._size:
            for i in range(index, self._size):
                self._array.set_unchecked(i, self._array.get_unchecked(i + 1))
        return value

    def insert(self, index: int, value: T):
        """Insert a copy of the given value at the given index.

        Preserves the relative order of the elements.

        Args:
            index: The index at which to insert the value. Must be in the range [0, size].
            value: The value to insert.
        """
        index = clamp(get_positive_index(index, self._size, include_end=True), 0, self._size)
        assert self._size < len(self._array), "Array is full"
        self._size += 1
        for i in range(self._size - 1, index, -1):
            self._array.set_unchecked(i, self._array.get_unchecked(i - 1))
        self._array.set_unchecked(index, value)

    def remove(self, value: T) -> bool:
        """Remove the first occurrence of the given value, returning whether the value was removed.

        Preserves the relative order of the elements.

        Args:
            value: The value to remove

        Returns:
            True if the value was removed, False otherwise.
        """
        index = self.index(value)
        if index < 0:
            return False
        self.pop(index)
        return True

    def clear(self):
        """Clear the array, removing all elements.

        References to elements are not immediately changed, but future insertions may overwrite them.
        """
        self._size = 0

    def set_add(self, value: T) -> bool:
        """Adds a copy of the given value if it is not already present, returning whether the value was added.

        If the value is already present, the array is not modified.
        If the array is full, the value is not added.

        Args:
            value: The value to add

        Returns:
            True if the value was added, False otherwise.
        """
        if self._size >= len(self._array):
            return False
        if value in self:
            return False
        self.append(value)
        return True

    def set_remove(self, value: T) -> bool:
        """Removes the first occurrence of the given value, returning whether the value was removed.

        Does not preserve the relative order of the elements.

        Args:
            value: The value to remove

        Returns:
            True if the value was removed, False otherwise.
        """
        index = self.index(value)
        if index < 0:
            return False
        if index < self._size - 1:
            self._array.set_unchecked(index, self._array.get_unchecked(self._size - 1))
        self._size -= 1
        return True

    def __iadd__(self, other):
        """Appends copies of the values in the given array to the end of the array."""
        self.extend(other)
        return self

    def __eq__(self, other):
        if not isinstance(other, ArrayLike):
            return False
        if len(self) != len(other):
            return False
        i = 0
        while i < len(self):
            if self.get_unchecked(i) != other.get_unchecked(i):
                return False
            i += 1
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        raise TypeError("unhashable type: 'VarArray'")


class ArrayPointer[T](Record, ArrayLike[T]):
    """An array defined by a size and pointer to the first element.

    Supports negative indexes.

    This is intended to be created internally and improper use may result in hard to debug issues.

    Usage:
        ```python
        ArrayPointer[T](size: int, block: int, offset: int)
        ```
    """

    size: int
    block: int
    offset: int

    def __len__(self) -> int:
        """Return the number of elements in the array."""
        return self.size

    @classmethod
    def element_type(cls) -> type[T]:
        """Return the type of the elements in the array."""
        return cls.type_var_value(T)

    @meta_fn
    def _get_item(self, item: int) -> T:
        if not ctx():
            raise TypeError("ArrayPointer values cannot be accessed outside of a context")
        return _deref(
            # Allows a compile time constant block so we can warn based on callback read/write access
            (self._value_["block"]._is_py_() and self._value_["block"]._as_py_()) or self.block,
            self.offset + Num._accept_(item) * Num._accept_(self.element_type()._size_()),
            self.element_type(),
        )

    def __getitem__(self, item) -> T:
        return self.get_unchecked(get_positive_index(item, self.size))

    def __setitem__(self, key: int, value: T):
        self.set_unchecked(get_positive_index(key, self.size), value)

    @meta_fn
    def get_unchecked(self, item: int) -> T:
        return self._get_item(item)._get_()

    @meta_fn
    def set_unchecked(self, key: int, value: T):
        dst = self._get_item(key)
        if self.element_type()._is_value_type_():
            dst._set_(value)
        else:
            dst._copy_from__(value)


class ArraySet[T, Capacity](Record):
    """A set implemented as an array with a fixed maximum capacity.

    Usage:
        ```python
        ArraySet[T, Capacity].new()  # Create a new empty set
        ```

    Examples:
        ```python
        s = ArraySet[int, 10].new()
        s.add(1)
        s.add(2)
        assert 1 in s
        assert 3 not in s
        s.remove(1)
        assert 1 not in s
        ```
    """

    _values: VarArray[T, Capacity]

    @classmethod
    def new(cls) -> Self:
        """Create a new empty set."""
        element_type = cls.type_var_value(T)
        capacity = cls.type_var_value(Capacity)
        return cls(VarArray[element_type, capacity].new())

    def __len__(self):
        """Return the number of elements in the set."""
        return len(self._values)

    def __contains__(self, value):
        """Return whether the given value is present in the set."""
        return value in self._values

    def __iter__(self) -> SonolusIterator[T]:
        """Return an iterator over the values in the set."""
        return self._values.__iter__()

    def add(self, value: T) -> bool:
        """Add a copy of the given value to the set.

        This has no effect and returns False if the value is already present or if the set is full.

        Args:
            value: The value to add.

        Returns:
            True if the value was added, False otherwise.
        """
        return self._values.set_add(value)

    def remove(self, value: T) -> bool:
        """Remove the given value from the set.

        This has no effect and returns False if the value is not present.

        Args:
            value: The value to remove.

        Returns:
            True if the value was removed, False otherwise.
        """
        return self._values.set_remove(value)

    def clear(self):
        """Clear the set, removing all elements."""
        self._values.clear()


class FrozenNumSet[Size](Record):
    _values: Array[Num, Size]

    @classmethod
    @meta_fn
    def of(cls, *values: Num) -> Self:
        if ctx():
            try:
                num_values = [Num._accept_(v) for v in values]
            except TypeError:
                raise TypeError("Only sets of numeric values are supported") from None
            if all(v._is_py_() for v in num_values):
                const_values = [v._as_py_() for v in num_values]
                arr = Array[Num, len(const_values)]._with_value([Num(v) for v in sorted(const_values)])
                return cls(arr)
            else:
                arr = Array[Num, len(values)](*values)
                compile_and_call(arr.sort)
        else:
            arr = Array[Num, len(values)](*sorted(values))
        return cls(arr)

    def __len__(self) -> int:
        return len(self._values)

    def __contains__(self, value: Num) -> bool:
        if len(self) < 8:
            return value in self._as_tuple()
        else:
            left = 0
            right = len(self) - 1
            while left <= right:
                mid = (left + right) // 2
                mid_value = self._values.get_unchecked(mid)
                if mid_value == value:
                    return True
                elif mid_value < value:
                    left = mid + 1
                else:
                    right = mid - 1
            return False

    def __iter__(self) -> SonolusIterator[Num]:
        return self._values.__iter__()

    @meta_fn
    def _as_tuple(self) -> tuple[Num, ...]:
        return tuple(self._values.get_unchecked(i) for i in range(Num._accept_(len(self))._as_py_()))


class _ArrayMapEntry[K, V](Record):
    key: K
    value: V


class ArrayMap[K, V, Capacity](Record):
    """A map implemented as an array of key-value pairs with a fixed maximum capacity.

    Usage:
        ```python
        ArrayMap[K, V, Capacity].new()  # Create a new empty map
        ```

    Examples:
        ```python
        map = ArrayMap[int, int, 10].new()
        map[1] = 2
        map[3] = 4
        assert 1 in map
        assert 2 not in map
        assert map[3] == 4
        ```
    """

    _size: int
    _array: Array[_ArrayMapEntry[K, V], Capacity]

    @classmethod
    def new(cls) -> Self:
        """Create a new empty map."""
        key_type = cls.type_var_value(K)
        value_type = cls.type_var_value(V)
        capacity = cls.type_var_value(Capacity)
        return cls(0, zeros(Array[_ArrayMapEntry[key_type, value_type], capacity]))

    def __len__(self) -> int:
        """Return the number of key-value pairs in the map."""
        return self._size

    @classmethod
    def capacity(cls) -> int:
        """Return the maximum number of key-value pairs the map can hold."""
        return cls.type_var_value(Capacity)

    def is_full(self) -> bool:
        """Return whether the map is full."""
        return self._size == self.capacity()

    def keys(self) -> SonolusIterator[K]:
        """Return an iterator over the keys in the map."""
        return _ArrayMapKeyIterator(self, 0)

    def values(self) -> SonolusIterator[V]:
        """Return an iterator over the values in the map."""
        return _ArrayMapValueIterator(self, 0)

    def items(self) -> SonolusIterator[tuple[K, V]]:
        """Return an iterator over the key-value pairs in the map."""
        return _ArrayMapEntryIterator(self, 0)

    def __iter__(self) -> SonolusIterator[K]:
        """Return an iterator over the keys in the map."""
        return self.keys()

    def __getitem__(self, key: K) -> V:
        """Return the value associated with the given key.

        Must be called with a key that is present in the map.

        The returned value continues to be part of the map.
        Future modifications to the map will affect the returned value.

        Notes:
            Future modifications to the map may cause unexpected changes to the returned value.
            If the map may be modified in the future, it's recommended to make a copy of the value.

            For example:
            ```python
            map = ArrayMap[int, Pair[int, int], 10].new()
            map[1] = Pair(2, 3)
            map[3] = Pair(4, 5)
            map[5] = Pair(6, 7)
            p = map[3]
            map.pop(1)
            # The value of `p` may now be different
            ```
        """
        for i in range(self._size):
            entry = self._array.get_unchecked(i)
            if entry.key == key:
                return entry.value
        error()

    def __setitem__(self, key: K, value: V):
        """Associate the given key with the given value.

        If the key is already present in the map, the value is updated.
        Must not be called if the map is full.

        Args:
            key: The key to associate with the value.
            value: The value to associate with the key
        """
        for i in range(self._size):
            entry = self._array.get_unchecked(i)
            if entry.key == key:
                entry.value = value
                return
        assert self._size < self.capacity(), "Map is full"
        self._array.set_unchecked(self._size, _ArrayMapEntry(key, value))
        self._size += 1

    def __delitem__(self, key: K):
        """Remove the key-value pair associated with the given key.

        Must be called with a key that is present in the map.

        Args:
            key: The key to remove
        """
        for i in range(self._size):
            entry = self._array.get_unchecked(i)
            if entry.key == key:
                self._size -= 1
                if i < self._size:
                    self._array.set_unchecked(i, self._array.get_unchecked(self._size))
                return
        error()

    def __contains__(self, key: K) -> bool:
        """Return whether the given key is present in the map.

        Args:
            key: The key to check for

        Returns:
            True if the key is present, False otherwise.
        """
        for i in range(self._size):  # noqa: SIM110
            if self._array.get_unchecked(i).key == key:
                return True
        return False

    def pop(self, key: K) -> V:
        """Remove and return a copy of the value associated with the given key.

        Must be called with a key that is present in the map.

        Args:
            key: The key to remove

        Returns:
            The value associated with the key
        """
        for i in range(self._size):
            entry = self._array.get_unchecked(i)
            if entry.key == key:
                value = copy(entry.value)
                self._size -= 1
                if i < self._size:
                    self._array.set_unchecked(i, self._array.get_unchecked(self._size))
                return value
        error()

    def clear(self):
        """Clear the map, removing all key-value pairs."""
        self._size = 0


class _ArrayMapKeyIterator[K, V, Capacity](Record, SonolusIterator):
    _map: ArrayMap[K, V, Capacity]
    _index: int

    def next(self) -> Maybe[K]:
        if self._index < len(self._map):
            key = self._map._array.get_unchecked(self._index).key
            self._index += 1
            return Some(key)
        return Nothing


class _ArrayMapValueIterator[K, V, Capacity](Record, SonolusIterator):
    _map: ArrayMap[K, V, Capacity]
    _index: int

    def next(self) -> Maybe[V]:
        if self._index < len(self._map):
            value = self._map._array.get_unchecked(self._index).value
            self._index += 1
            return Some(value)
        return Nothing


class _ArrayMapEntryIterator[K, V, Capacity](Record, SonolusIterator):
    _map: ArrayMap[K, V, Capacity]
    _index: int

    def next(self) -> Maybe[tuple[K, V]]:
        if self._index < len(self._map):
            entry = self._map._array.get_unchecked(self._index)
            result = (entry.key, entry.value)
            self._index += 1
            return Some(result)
        return Nothing


class _LinkedListNodeRef[TKey, TValue](Protocol):
    def get_value(self) -> TValue: ...

    def get_next(self) -> Self: ...

    def set_next(self, next_node: Self): ...

    def set_prev(self, prev_node: Self):
        # No-op for singly linked lists
        return

    def is_present(self) -> bool: ...

    def set(self, other: Self): ...

    def copy(self) -> Self: ...

    def empty(self) -> Self: ...


def _merge_linked_list_nodes[TNode: _LinkedListNodeRef](
    a: TNode,
    b: TNode,
) -> TNode:
    head = a.empty()
    tail = a.empty()
    left = a.copy()
    right = b.copy()

    while left.is_present() and right.is_present():
        if left.get_value() <= right.get_value():
            if not head.is_present():
                head.set(left)
                tail.set(left)
            else:
                tail.set_next(left)
                tail.set(left)
            left.set(left.get_next())
        else:
            if not head.is_present():
                head.set(right)
                tail.set(right)
            else:
                tail.set_next(right)
                tail.set(right)
            right.set(right.get_next())

    while left.is_present():
        if not head.is_present():
            head.set(left)
            tail.set(left)
        else:
            tail.set_next(left)
            tail.set(left)
        left.set(left.get_next())

    while right.is_present():
        if not head.is_present():
            head.set(right)
            tail.set(right)
        else:
            tail.set_next(right)
            tail.set(right)
        right.set(right.get_next())

    if tail.is_present():
        tail.set_next(a.empty())

    return head


def _merge_sort_linked_list_nodes[TNode: _LinkedListNodeRef](
    head: TNode,
) -> TNode:
    # Calculate length
    length = 0
    node = head.copy()
    while node.is_present():
        length += 1
        node.set(node.get_next())

    # Trivial case
    if length <= 1:
        return head

    # Bottom-up merge sort: start with sublists of size 1, then 2, 4, 8, etc.
    size = 1
    while size < length:
        current = head.copy()
        new_head = head.empty()
        new_tail = head.empty()

        # Process all pairs of sublists of the current size
        while current.is_present():
            # Extract the first sublist
            left = current.copy()
            prev = current.empty()
            i = 0
            while i < size and current.is_present():
                prev.set(current)
                current.set(current.get_next())
                i += 1
            if prev.is_present():
                prev.set_next(prev.empty())

            # We've made it to the end without a second sublist to merge, so just attach it to the end
            if not current.is_present():
                # Since size < length, we know a full iteration must have happened already, so new_tail is valid
                new_tail.set_next(left)
                break

            # Extract the second sublist
            right = current.copy()
            prev = current.empty()
            i = 0
            while i < size and current.is_present():
                prev.set(current)
                current.set(current.get_next())
                i += 1
            if prev.is_present():
                prev.set_next(prev.empty())

            merged = _merge_linked_list_nodes(left, right)

            # Append the merged result
            if not new_head.is_present():
                new_head.set(merged)
                new_tail.set(merged)
            else:
                new_tail.set_next(merged)

            # Move tail to the end of the merged section
            while new_tail.get_next().is_present():
                new_tail.set(new_tail.get_next())

        # Update head for the next iteration
        head.set(new_head)
        size *= 2

    return head


class _EntityNodeRef[Archetype, GetValue, GetNextRef, GetPrevRef](Record):
    index: int

    def get_value(self) -> Any:
        return self.get_value_fn(self.archetype.at(self.index))

    def get_next(self) -> _EntityNodeRef:
        next_ref = self.get_next_ref_fn(self.archetype.at(self.index))
        return self.with_index(next_ref.index)

    def set_next(self, next_node: _EntityNodeRef):
        entity = self.archetype.at(self.index)
        next_ref = self.get_next_ref_fn(entity)
        next_ref.index = next_node.index

    def set_prev(self, prev_node: _EntityNodeRef):
        if self.get_prev_ref_fn is not None:
            entity = self.archetype.at(self.index)
            prev_ref = self.get_prev_ref_fn(entity)
            prev_ref.index = prev_node.index

    def is_present(self) -> bool:
        return self.index > 0

    def set(self, other: _EntityNodeRef):
        self.index = other.index

    def copy(self) -> _EntityNodeRef:
        return self.with_index(self.index)

    def empty(self) -> _EntityNodeRef:
        return self.with_index(0)

    def with_index(self, index: int) -> _EntityNodeRef:
        return _EntityNodeRef[
            self.archetype,
            self.get_value_fn,
            self.get_next_ref_fn,
            self.get_prev_ref_fn,
        ](index)

    @property
    def archetype(self):
        return self.type_var_value(Archetype)

    @property
    def get_value_fn(self):
        return self.type_var_value(GetValue)

    @property
    def get_next_ref_fn(self):
        return self.type_var_value(GetNextRef)

    @property
    def get_prev_ref_fn(self):
        return self.type_var_value(GetPrevRef)


def sort_linked_entities[T: AnyArchetype](
    head_ref: EntityRef[T],
    /,
    *,
    get_value: Callable[[T], Any],
    get_next_ref: Callable[[T], EntityRef[T]],
    get_prev_ref: Callable[[T], EntityRef[T]] | None = None,
) -> EntityRef[T]:
    """Sort a linked list of entities using merge sort.

    If get_prev_ref is provided, the backward links will be updated as well.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            sort_key: int
            next: EntityRef[MyArchetype]

        def sort_my_archetype(head: EntityRef[MyArchetype]) -> EntityRef[MyArchetype]:
            return sort_linked_entities(
                head,
                get_value=lambda e: e.sort_key,
                get_next_ref=lambda e: e.next,
            )
        ```

    Args:
        head_ref: A reference to the head of the linked list.
        get_value: A function that takes an entity and returns the value to sort by.
        get_next_ref: A function that takes an entity and returns a reference to the next entity.
        get_prev_ref: An optional function that takes an entity and returns a reference to the previous entity.

    Returns:
        A reference to the head of the sorted linked list.
    """
    archetype = head_ref.archetype()

    sorted_head_index = _merge_sort_linked_list_nodes(
        _EntityNodeRef[archetype, get_value, get_next_ref, get_prev_ref](head_ref.index)
    ).index

    if get_prev_ref is not None:
        current_ref = _EntityNodeRef[archetype, get_value, get_next_ref, get_prev_ref](sorted_head_index)
        prev_ref = current_ref.empty()
        while current_ref.is_present():
            current_ref.set_prev(prev_ref)
            prev_ref.set(current_ref)
            current_ref.set(current_ref.get_next())

    return EntityRef[archetype](sorted_head_index)



from __future__ import annotations

import argparse
import contextlib
import gc
import http.server
import importlib
import queue
import shlex
import shutil
import socket
import socketserver
import sys
import textwrap
import threading
import traceback
from dataclasses import dataclass
from pathlib import Path
from time import perf_counter
from typing import TYPE_CHECKING, NamedTuple, Protocol

from sonolus.backend.excepthook import print_simple_traceback
from sonolus.backend.utils import get_function, get_functions, get_tree_from_file
from sonolus.build.compile import CompileCache
from sonolus.script.internal.context import ProjectContextState
from sonolus.script.internal.error import CompilationError

if TYPE_CHECKING:
    from sonolus.script.project import BuildConfig, Project

HELP_TEXT = """
[r]ebuild
[d]ecode <message_code>
[h]elp
[q]uit
""".strip()

HELP_TEXT = textwrap.dedent(HELP_TEXT)


class CommandHelp(NamedTuple):
    alias: str
    command: str
    description: list[str]


DETAILED_HELP_TEXT = [
    CommandHelp(
        alias="r",
        command="rebuild",
        description=[
            ("Rebuild the project with the latest changes from source files."),
        ],
    ),
    CommandHelp(
        alias="d",
        command="decode <message_code>",
        description=[
            (
                "Decode a debug message code to its original text with a stack trace. "
                "Message codes are generated by assert statements and certain debug-related functions such as "
                "debug.error() and debug.notify(). "
                "The game is automatically paused when these are triggered in debug mode and the message code "
                "can be found in the debug log."
            ),
            "Example: d 42",
        ],
    ),
    CommandHelp(
        alias="h",
        command="help",
        description=[
            "Show this help message.",
        ],
    ),
    CommandHelp(
        alias="q",
        command="quit",
        description=[
            "Exit the development server.",
        ],
    ),
]


@dataclass
class ServerState:
    project: Project
    project_module_name: str
    core_module_names: set[str]
    build_dir: Path
    config: BuildConfig
    cache: CompileCache
    project_state: ProjectContextState


class Command(Protocol):
    def execute(self, server_state: ServerState) -> None: ...


@dataclass
class RebuildCommand:
    def execute(self, server_state: ServerState):
        from sonolus.build.cli import build_collection

        for module_name in tuple(sys.modules):
            if module_name not in server_state.core_module_names:
                del sys.modules[module_name]

        try:
            project_module = importlib.import_module(server_state.project_module_name)
        except Exception:
            print(traceback.format_exc())
            return

        get_function.cache_clear()
        get_tree_from_file.cache_clear()
        get_functions.cache_clear()
        print("Rebuilding...")
        try:
            start_time = perf_counter()
            server_state.cache.reset_accessed()
            server_state.project_state = ProjectContextState.from_build_config(server_state.config)
            server_state.project = project_module.project
            build_collection(
                server_state.project,
                server_state.build_dir,
                server_state.config,
                cache=server_state.cache,
                project_state=server_state.project_state,
            )
            server_state.cache.prune_unaccessed()
            end_time = perf_counter()
            print(f"Rebuild completed in {end_time - start_time:.2f} seconds")
        except CompilationError:
            exc_info = sys.exc_info()
            if server_state.config.verbose:
                print(traceback.format_exc())
            else:
                print_simple_traceback(*exc_info)
                print("\nFor more details, run with the --verbose (-v) flag.")


@dataclass
class DecodeCommand:
    message_code: int

    def execute(self, server_state: ServerState):
        debug_str_mappings = server_state.project_state.debug_str_mappings
        message = next((msg for msg, code in debug_str_mappings.items() if code == self.message_code), None)

        if message is not None:
            print(message)
        else:
            print(f"Unknown message code: {self.message_code}")


@dataclass
class HelpCommand:
    def execute(self, server_state: ServerState):
        terminal_width = shutil.get_terminal_size().columns
        max_width = min(terminal_width, 120)

        print("Available Commands:\n")

        for entry in DETAILED_HELP_TEXT:
            print(f"[{entry.alias}] {entry.command}")

            for paragraph in entry.description:
                initial_indent = "  "
                subsequent_indent = "  "
                wrapped = textwrap.fill(
                    paragraph,
                    width=max_width - len(initial_indent),
                    initial_indent=initial_indent,
                    subsequent_indent=subsequent_indent,
                )
                print(wrapped)
                print()


@dataclass
class ExitCommand:
    def execute(self, server_state: ServerState):
        print("Exiting...")
        sys.exit(0)


def parse_dev_command(command_line: str) -> Command | None:
    parser = argparse.ArgumentParser(prog="", add_help=False, exit_on_error=False)
    subparsers = parser.add_subparsers(dest="cmd")

    subparsers.add_parser("rebuild", aliases=["r"])
    decode_parser = subparsers.add_parser("decode", aliases=["d"])
    decode_parser.add_argument("message_code", type=int, help="Message code to decode")
    subparsers.add_parser("help", aliases=["h"])
    subparsers.add_parser("quit", aliases=["q"])

    try:
        split_args = shlex.split(command_line)
    except ValueError as e:
        print(f"Error parsing command: {e}\n")
        return None

    try:
        args = parser.parse_args(split_args)
        if args.cmd in {"rebuild", "r"}:
            return RebuildCommand()
        elif args.cmd in {"decode", "d"}:
            return DecodeCommand(message_code=args.message_code)
        elif args.cmd in {"help", "h"}:
            return HelpCommand()
        elif args.cmd in {"quit", "q"}:
            return ExitCommand()
        else:
            # Really, we should not reach here, since argparse would have errored out earlier
            print("Unknown command.\n")
            return None
    except (argparse.ArgumentError, argparse.ArgumentTypeError) as e:
        print(f"Error parsing command: {e}\n")
        return None
    except SystemExit:
        # argparse throws this on some errors, and will print out help automatically
        print()
        return None


def command_input_thread(command_queue: queue.Queue, prompt_event: threading.Event):
    print(f"\nAvailable commands:\n{HELP_TEXT}")

    while True:
        try:
            prompt_event.wait()
            prompt_event.clear()

            print("\n> ", end="", flush=True)
            command_line = input()
            if command_line.strip():
                cmd = parse_dev_command(command_line.strip())
                if cmd:
                    command_queue.put(cmd)
                    if isinstance(cmd, ExitCommand):
                        break
                else:
                    print(f"Available commands:\n{HELP_TEXT}")
                    # Show prompt again
                    prompt_event.set()
            else:
                prompt_event.set()
        except EOFError:
            break
        except Exception as e:
            print(f"Error reading command: {e}\n")
            prompt_event.set()


def get_local_ips():
    hostname = socket.gethostname()
    local_ips = []

    with contextlib.suppress(socket.gaierror):
        local_ips.append(socket.gethostbyname(socket.getfqdn()))

    try:
        for info in socket.getaddrinfo(hostname, None):
            ip = info[4][0]
            if not ip.startswith("127.") and ":" not in ip:
                local_ips.append(ip)
    except socket.gaierror:
        pass

    return sorted(set(local_ips))


def run_server(
    base_dir: Path,
    port: int,
    project_module_name: str | None,
    core_module_names: set[str] | None,
    build_dir: Path,
    config: BuildConfig,
    project,
):
    from sonolus.build.cli import build_collection

    cache = CompileCache()
    project_state = ProjectContextState.from_build_config(config)

    start_time = perf_counter()
    build_collection(project, build_dir, config, cache=cache, project_state=project_state)
    end_time = perf_counter()
    print(f"Build finished in {end_time - start_time:.2f}s")

    interactive = project_module_name is not None and core_module_names is not None

    class DirectoryHandler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=str(base_dir), **kwargs)

        def log_message(self, fmt, *args):
            sys.stdout.write("\r\033[K")  # Clear line
            sys.stdout.write(f"{self.address_string()} [{self.log_date_time_string()}] {fmt % args}\n")
            if interactive:
                sys.stdout.write("> ")
            sys.stdout.flush()

    with socketserver.TCPServer(("", port), DirectoryHandler) as httpd:
        local_ips = get_local_ips()
        print(f"Server started on port {port}")
        print("Available on:")
        for ip in local_ips:
            print(f"  http://{ip}:{port}")

        if interactive:
            server_state = ServerState(
                project=project,
                project_module_name=project_module_name,
                core_module_names=core_module_names,
                build_dir=build_dir,
                config=config,
                cache=cache,
                project_state=project_state,
            )

            threading.Thread(target=httpd.serve_forever, daemon=True).start()

            command_queue = queue.Queue()
            prompt_event = threading.Event()
            input_thread = threading.Thread(
                target=command_input_thread, args=(command_queue, prompt_event), daemon=True
            )
            input_thread.start()

            prompt_event.set()

            try:
                while True:
                    cmd = command_queue.get()
                    try:
                        cmd.execute(server_state)
                        gc.collect()
                    except Exception:
                        print(f"{traceback.format_exc()}\n")
                    prompt_event.set()
            except KeyboardInterrupt:
                print("Exiting...")
                sys.exit(0)
            finally:
                httpd.shutdown()
        else:
            httpd.serve_forever()



from pydori.lib.skin import Skin
from pydori.preview.connector import PreviewHoldConnector, PreviewSimLine
from pydori.preview.event import PreviewBpmChange, PreviewTimescaleChange
from pydori.preview.note import ALL_PREVIEW_NOTE_TYPES
from pydori.preview.stage import PreviewStage
from sonolus.script.engine import PreviewMode

preview_mode = PreviewMode(
    archetypes=[
        PreviewStage,
        *ALL_PREVIEW_NOTE_TYPES,
        PreviewHoldConnector,
        PreviewSimLine,
        PreviewBpmChange,
        PreviewTimescaleChange,
    ],
    skin=Skin,
)



from sonolus.backend.ops import Op
from sonolus.backend.place import BlockPlace, Place, SSAPlace

type IRExpr = IRConst | IRPureInstr | IRGet
type IRStmt = IRExpr | IRInstr | IRSet


_IR_CONST_CACHE_START = -5
_IR_CONST_CACHE_STOP = 257


class IRConst:
    __slots__ = ("value",)

    value: float | int

    def __new__(cls, value):
        if float(value).is_integer():
            int_value = int(value)
            if _IR_CONST_CACHE_START <= int_value < _IR_CONST_CACHE_STOP:
                return _IR_CONST_CACHE[int_value - _IR_CONST_CACHE_START]
            else:
                return _create_raw_const(int_value)
        return super().__new__(cls)

    def __init__(self, value: float):
        self.value = value

    def __repr__(self):
        return f"IRConst({self.value!r})"

    def __str__(self):
        return f"{self.value}"

    def __eq__(self, other):
        return isinstance(other, IRConst) and self.value == other.value

    def __hash__(self):
        return hash(self.value)


def _create_raw_const(value: float | int) -> IRConst:
    result = object.__new__(IRConst)
    result.value = value
    return result


_IR_CONST_CACHE = tuple(_create_raw_const(i) for i in range(_IR_CONST_CACHE_START, _IR_CONST_CACHE_STOP))


class IRPureInstr:
    __slots__ = ("args", "array_defs", "defs", "is_array_init", "live", "op", "uses", "visited")

    op: Op
    args: list[IRExpr]

    def __init__(self, op: Op, args: list[IRExpr]):
        assert op.pure
        self.op = op
        self.args = args

    def __repr__(self):
        return f"IRPureInstr({self.op!r}, {self.args!r})"

    def __str__(self):
        return f"{self.op.name}({', '.join(map(str, self.args))})"

    def __eq__(self, other):
        return isinstance(other, IRPureInstr) and self.op == other.op and self.args == other.args

    def __hash__(self):
        return hash((self.op, tuple(self.args)))


class IRInstr:
    __slots__ = ("args", "array_defs", "defs", "is_array_init", "live", "op", "uses", "visited")

    op: Op
    args: list[IRExpr]

    def __init__(self, op: Op, args: list[IRExpr]):
        self.op = op
        self.args = args

    def __repr__(self):
        return f"IRInstr({self.op!r}, {self.args!r})"

    def __str__(self):
        return f"{self.op.name}({', '.join(map(str, self.args))})"

    def __eq__(self, other):
        return isinstance(other, IRInstr) and self.op == other.op and self.args == other.args

    def __hash__(self):
        return hash((self.op, tuple(self.args)))


class IRGet:
    __slots__ = ("array_defs", "defs", "is_array_init", "live", "place", "uses", "visited")

    place: Place

    def __init__(self, place: Place):
        self.place = place

    def __repr__(self):
        return f"IRGet({self.place!r})"

    def __str__(self):
        return f"{self.place}"

    def __eq__(self, other):
        return isinstance(other, IRGet) and self.place == other.place

    def __hash__(self):
        return hash(self.place)


class IRSet:
    __slots__ = ("array_defs", "defs", "is_array_init", "live", "place", "uses", "value", "visited")

    place: Place
    value: IRExpr | IRInstr

    def __init__(self, place: Place, value: IRExpr | IRInstr):
        self.place = place
        self.value = value

    def __repr__(self):
        return f"IRSet({self.place!r}, {self.value!r})"

    def __str__(self):
        match self.place:
            case BlockPlace():
                return f"{self.place} <- {self.value}"
            case SSAPlace():
                return f"{self.place} := {self.value}"
            case _:
                raise TypeError(f"Invalid place: {self.place}")

    def __eq__(self, other):
        return isinstance(other, IRSet) and self.place == other.place and self.value == other.value

    def __hash__(self):
        return hash((self.place, self.value))



from collections.abc import Callable, Sequence
from contextvars import ContextVar
from typing import Any, Literal, Never, assert_never

from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import cfg_to_mermaid
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig, run_passes
from sonolus.backend.optimize.simplify import RenumberVars
from sonolus.script.internal.context import ModeContextState, ProjectContextState, RuntimeChecks, ctx, set_ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.native import native_function
from sonolus.script.internal.simulation_context import SimulationContext
from sonolus.script.num import Num

debug_log_callback = ContextVar[Callable[[Num], None]]("debug_log_callback")


@meta_fn
def error(message: str | None = None) -> Never:  # type: ignore
    """Raise an error, and if runtime checks are set to notify, log a message and pause the game.

    This function is used to raise an error during runtime and terminate the current callback.

    If runtime checks are set to notify (default in dev), this function will log a message and pause the game
    before terminating.

    Args:
        message: The message to log.
    """
    message = validate_value(message)._as_py_() or "Error"  # type: ignore
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    if ctx():
        match ctx().project_state.runtime_checks:
            case RuntimeChecks.NOTIFY_AND_TERMINATE:
                debug_log(ctx().map_debug_message(message))
                debug_pause()
            case RuntimeChecks.TERMINATE | RuntimeChecks.NONE:
                pass
            case _ as unreachable:
                assert_never(unreachable)
        terminate()
    else:
        raise RuntimeError(message)


@meta_fn
def static_error(message: str | None = None) -> Never:
    """Raise a static error.

    This function is used to raise an error during compile-time if the compiler cannot guarantee that
    this function will not be called during runtime.

    Args:
        message: The message to log.
    """
    message = validate_value(message)._as_py_() or "Error"  # type: ignore
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    raise RuntimeError(message)


@meta_fn
def debug_log(value: int | float | bool):
    """Log a value in debug mode."""
    if debug_log_callback.get(None):
        return debug_log_callback.get()(value)  # type: ignore
    else:
        return _debug_log(value)


@native_function(Op.DebugLog)
def _debug_log(value: int | float | bool):
    print(f"[DEBUG] {value}")
    return 0


@native_function(Op.DebugPause)
def debug_pause():
    """Pause the game if in debug mode."""
    input("[DEBUG] Paused")


@meta_fn
def notify(message: str):
    """Log a code that can be decoded by the dev server and pause the game if runtime checks are set to notify.

    If runtime checks are not set to notify, this function will do nothing.

    Args:
        message: The message to log.
    """
    message = validate_value(message)._as_py_()  # type: ignore
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    if ctx():
        if ctx().project_state.runtime_checks == RuntimeChecks.NOTIFY_AND_TERMINATE:
            debug_log(ctx().map_debug_message(message))
            debug_pause()
    else:
        print(f"[NOTIFY] {message}")


@meta_fn
def runtime_checks_enabled() -> bool:
    if ctx():
        return ctx().project_state.runtime_checks != RuntimeChecks.NONE
    else:
        return True


@meta_fn
def require(value: int | float | bool, message: str | None = None):
    """Require a condition to be true, or raise an error.

    Similar to assert, but will terminate the current callback even if runtime checks are set to none.

    If runtime checks are set to notify (default in dev), this function will log a message and pause the game
    before terminating.

    Args:
        value: The condition to check.
        message: The message to log if the condition is false.
    """
    if not ctx():
        if not value:
            raise AssertionError(message if message is not None else "Assertion failed")
        return
    value = Num._accept_(validate_value(value))
    message = validate_value(message)
    message = message._as_py_() or "Assertion failed"
    if value._is_py_():
        if value._as_py_():
            return
        else:
            error(message)
    else:
        ctx().test = value.ir()
        t_branch = ctx().branch(None)
        f_branch = ctx().branch(0)
        set_ctx(f_branch)
        error(message)  # type: ignore
        set_ctx(t_branch)


@meta_fn
def assert_true(value: int | float | bool, message: str | None = None):
    value = validate_value(value)
    if (
        ctx()
        and ctx().project_state.runtime_checks == RuntimeChecks.NONE
        and not (validate_value(value)._is_py_() and validate_value(value)._as_py_() == 0)
    ):
        # Don't do anything if runtime checks are disabled, unless the value is statically known to be false.
        # `assert False` is a somewhat common pattern to indicate unreachable code, and stripping that out can
        # cause compilation errors, so we retain those checks.
        return
    require(value, message)


@meta_fn
def assert_false(value: int | float | bool, message: str | None = None):
    assert_true(value == 0, message)


def static_assert(value: int | float | bool, message: str | None = None):
    message = message if message is not None else "Static assertion failed"
    if not _is_static_true(value):
        static_error(message)


def try_static_assert(value: int | float | bool, message: str | None = None):
    message = message if message is not None else "Static assertion failed"
    if _is_static_false(value):
        static_error(message)
    if not value:
        error(message)


@meta_fn
def assert_unreachable(message: str | None = None) -> Never:
    # This works a bit differently from assert_never from typing in that it throws an error if the Sonolus.py
    # compiler cannot guarantee that this function will not be called, which is different from what type checkers
    # may be able to infer.
    message = validate_value(message)._as_py_() or "Unreachable code reached"  # type: ignore
    raise RuntimeError(message)


@meta_fn
def terminate():
    if ctx():
        set_ctx(ctx().into_dead())
    else:
        raise RuntimeError("Terminated")


def visualize_cfg(
    fn: Callable[[], Any] | Callable[[type], Any],
    /,
    *,
    mode: Mode = Mode.PLAY,
    callback: str = "",
    archetype: type | None = None,
    archetypes: list[type] | None = None,
    passes: Sequence[CompilerPass] | Literal["minimal", "fast", "standard"] = "fast",
) -> str:
    from sonolus.backend.optimize.optimize import FAST_PASSES, MINIMAL_PASSES, STANDARD_PASSES
    from sonolus.build.compile import callback_to_cfg

    match passes:
        case "minimal":
            passes = [
                *MINIMAL_PASSES[:-1],
                RenumberVars(),
            ]
        case "fast":
            passes = [
                *FAST_PASSES[:-1],
                RenumberVars(),
            ]
        case "standard":
            passes = [
                *STANDARD_PASSES[:-1],
                RenumberVars(),
            ]

    project_state = ProjectContextState()
    mode_state = ModeContextState(
        mode,
        archetypes,
    )

    cfg = callback_to_cfg(project_state, mode_state, fn, callback, archetype=archetype)  # type: ignore
    cfg = run_passes(cfg, passes, OptimizerConfig(mode=mode))
    return cfg_to_mermaid(cfg)


def simulation_context() -> SimulationContext:
    return SimulationContext()


@meta_fn
def _is_static_true(value: int | float | bool) -> bool:
    if ctx() is None:
        return bool(value)
    else:
        value = validate_value(value)
        return value._is_py_() and value._as_py_()


@meta_fn
def _is_static_false(value: int | float | bool) -> bool:
    if ctx() is None:
        return not bool(value)
    else:
        value = validate_value(value)
        return value._is_py_() and not value._as_py_()



from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet, IRStmt
from sonolus.backend.optimize.dominance import DominanceFrontiers, get_df, get_dom_children
from sonolus.backend.optimize.flow import BasicBlock, FlowEdge, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock


class ToSSA(CompilerPass):
    def requires(self) -> set[CompilerPass]:
        return {DominanceFrontiers()}

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        defs = self.defs_to_blocks(entry)
        self.insert_phis(defs)
        self.rename(entry, defs, {var: [] for var in defs}, {})
        self.remove_placeholder_phis(entry)
        return entry

    def rename(
        self,
        block: BasicBlock,
        defs: dict[TempBlock, set[BasicBlock]],
        ssa_places: dict[TempBlock, list[SSAPlace]],
        used: dict[str, int],
    ):
        to_pop = []
        for var, args in [*block.phis.items()]:
            if isinstance(var, SSAPlace):
                continue
            ssa_places[var].append(self.get_new_ssa_place(var.name, used))
            to_pop.append(var)
            block.phis[ssa_places[var][-1]] = args
        block.statements = [self.rename_stmt(stmt, ssa_places, used, to_pop) for stmt in block.statements]
        for edge in block.outgoing:
            dst = edge.dst
            for var, args in dst.phis.items():
                if isinstance(var, SSAPlace):
                    continue
                if ssa_places[var]:
                    args[block] = ssa_places[var][-1]
        block.test = self.rename_stmt(block.test, ssa_places, used, to_pop)
        for dom_child in get_dom_children(block):
            self.rename(dom_child, defs, ssa_places, used)
        for var in to_pop:
            ssa_places[var].pop()

    def remove_placeholder_phis(self, entry: BasicBlock):
        for block in traverse_cfg_preorder(entry):
            block.phis = {var: args for var, args in block.phis.items() if isinstance(var, SSAPlace)}

    def rename_stmt(
        self, stmt: IRStmt, ssa_places: dict[TempBlock, list[SSAPlace]], used: dict[str, int], to_pop: list[SSAPlace]
    ):
        match stmt:
            case IRConst():
                return stmt
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.rename_stmt(arg, ssa_places, used, to_pop) for arg in args])
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.rename_stmt(arg, ssa_places, used, to_pop) for arg in args])
            case IRGet(place=place):
                return IRGet(place=self.rename_stmt(place, ssa_places, used, to_pop))
            case IRSet(place=place, value=value):
                value = self.rename_stmt(value, ssa_places, used, to_pop)
                if isinstance(place, BlockPlace) and isinstance(place.block, TempBlock) and place.block.size == 1:
                    ssa_places[place.block].append(self.get_new_ssa_place(place.block.name, used))
                    to_pop.append(place.block)
                place = self.rename_stmt(place, ssa_places, used, to_pop)
                return IRSet(place=place, value=value)
            case SSAPlace():
                return stmt
            case TempBlock() if stmt.size == 1:
                if stmt not in ssa_places or not ssa_places[stmt]:
                    # This is an access to a definitely undefined variable
                    # But it might not be reachable in reality, so we should allow it
                    # Maybe there should be an error if this still happens after optimization,
                    # but recovering the location of the error in the original code is hard.
                    # This can happen in places like matching a VarArray[Num, 1] which was just created.
                    # IR generation won't immediately fold a check that size > 0 to false, so here we
                    # might see an access to uninitialized memory even though it's not reachable in reality.
                    return SSAPlace("err", 0)
                return ssa_places[stmt][-1]
            case TempBlock():
                return stmt
            case int():
                return stmt
            case BlockPlace(block=block, index=index, offset=offset):
                if isinstance(block, TempBlock) and block.size == 1:
                    return self.rename_stmt(block, ssa_places, used, to_pop)
                return BlockPlace(
                    block=self.rename_stmt(block, ssa_places, used, to_pop),
                    index=self.rename_stmt(index, ssa_places, used, to_pop),
                    offset=self.rename_stmt(offset, ssa_places, used, to_pop),
                )
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def insert_phis(self, defs: dict[TempBlock, set[BasicBlock]]):
        for var, blocks in defs.items():
            df = self.get_iterated_df(blocks)
            for block in df:
                block.phis[var] = {}

    def defs_to_blocks(self, entry: BasicBlock) -> dict[TempBlock, set[BasicBlock]]:
        result = {}
        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                def_block = self.get_stmt_def(stmt)
                if def_block is not None:
                    result.setdefault(def_block, set()).add(block)
        return result

    def get_stmt_def(self, stmt: IRStmt) -> TempBlock:
        if (
            isinstance(stmt, IRSet)
            and isinstance(stmt.place, BlockPlace)
            and isinstance(stmt.place.block, TempBlock)
            and stmt.place.block.size == 1
        ):
            return stmt.place.block
        return None

    def get_iterated_df(self, blocks: set[BasicBlock]) -> set[BasicBlock]:
        df = set()
        worklist = set(blocks)
        while worklist:
            block = worklist.pop()
            new_df = get_df(block) - df
            if new_df:
                df.update(new_df)
                worklist.update(new_df)
        return df

    def get_new_ssa_place(self, name: str, used: dict[str, int]) -> SSAPlace:
        if name not in used:
            used[name] = 0
        used[name] += 1
        return SSAPlace(name, used[name])


class FromSSA(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        for block in [*traverse_cfg_preorder(entry)]:
            self.process_block(block)
        return entry

    def process_block(self, block: BasicBlock):
        orig_incoming = [*block.incoming]
        block.incoming.clear()
        for edge in orig_incoming:
            between_block = BasicBlock()
            edge.dst = between_block
            between_block.incoming.add(edge)
            next_edge = FlowEdge(between_block, block, None)
            block.incoming.add(next_edge)
            between_block.outgoing.add(next_edge)
            for args in block.phis.values():
                if edge.src in args:
                    args[between_block] = args[edge.src]
        for edge in orig_incoming:
            # Multiple edges with different conditions can connect two of the same blocks,
            # so we need to remove the old phi arguments in a pass at the end.
            for args in block.phis.values():
                if edge.src in args:
                    del args[edge.src]
        incoming_blocks = {edge.src for edge in block.incoming}
        args_by_src = {}
        for args in block.phis.values():
            for src, arg in args.items():
                if src not in args_by_src:
                    args_by_src[src] = set()
                args_by_src[src].add(arg)
        for var, args in block.phis.items():
            for src, arg in args.items():
                if src not in incoming_blocks:
                    # Edges may have been rewritten so a phi refers to a block that is no longer directly connected.
                    continue
                if var in args_by_src[src]:
                    # Make an extra copy first of values that may be overwritten by another assignment.
                    src.statements.append(
                        IRSet(
                            place=self.place_from_ssa_place(var, "*"), value=IRGet(place=self.place_from_ssa_place(arg))
                        )
                    )
        for var, args in block.phis.items():
            for src, arg in args.items():
                if src not in incoming_blocks:
                    continue
                if var in args_by_src[src]:
                    src.statements.append(
                        IRSet(
                            place=self.place_from_ssa_place(var), value=IRGet(place=self.place_from_ssa_place(var, "*"))
                        )
                    )
                else:
                    src.statements.append(
                        IRSet(place=self.place_from_ssa_place(var), value=IRGet(place=self.place_from_ssa_place(arg)))
                    )
        block.phis = {}
        block.statements = [self.process_stmt(stmt) for stmt in block.statements]
        block.test = self.process_stmt(block.test)

    def process_stmt(self, stmt: IRStmt):
        match stmt:
            case IRConst():
                return stmt
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.process_stmt(arg) for arg in args])
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.process_stmt(arg) for arg in args])
            case IRGet(place=place):
                return IRGet(place=self.process_stmt(place))
            case IRSet(place=place, value=value):
                return IRSet(place=self.process_stmt(place), value=self.process_stmt(value))
            case SSAPlace():
                return self.place_from_ssa_place(stmt)
            case TempBlock():
                return stmt
            case int():
                return stmt
            case BlockPlace(block=block, index=index, offset=offset):
                return BlockPlace(
                    block=self.process_stmt(block),
                    index=self.process_stmt(index),
                    offset=self.process_stmt(offset),
                )
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def temp_block_from_ssa_place(self, ssa_place: SSAPlace, suffix: str = "") -> TempBlock:
        return TempBlock(f"{ssa_place.name}.{ssa_place.num}{suffix}")

    def place_from_ssa_place(self, ssa_place: SSAPlace, suffix: str = "") -> BlockPlace:
        return BlockPlace(block=self.temp_block_from_ssa_place(ssa_place, suffix), index=0, offset=0)



from pydori.lib.buckets import Buckets
from pydori.lib.effect import Effects
from pydori.lib.particle import Particles
from pydori.lib.skin import Skin
from pydori.watch.connector import WatchHoldConnector, WatchSimLine
from pydori.watch.event import WatchBpmChange, WatchTimescaleChange
from pydori.watch.note import ALL_WATCH_NOTE_TYPES, WatchHoldManager
from pydori.watch.stage import WatchScheduledLaneEffect, WatchStage
from pydori.watch.update_spawn import update_spawn
from sonolus.script.engine import WatchMode

watch_mode = WatchMode(
    archetypes=[
        WatchStage,
        WatchScheduledLaneEffect,
        *ALL_WATCH_NOTE_TYPES,
        WatchHoldManager,
        WatchHoldConnector,
        WatchSimLine,
        WatchBpmChange,
        WatchTimescaleChange,
    ],
    skin=Skin,
    effects=Effects,
    particles=Particles,
    buckets=Buckets,
    update_spawn=update_spawn,
)



from collections.abc import Iterator

from sonolus.script.array import Dim
from sonolus.script.containers import ArraySet
from sonolus.script.globals import level_memory
from sonolus.script.runtime import Touch, touches


@level_memory
class InputState:
    claimed_touches: ArraySet[int, Dim[16]]


def refresh_input_state():
    """Refresh the input data at the start of each frame."""
    InputState.claimed_touches.clear()


def claim_touch(touch_id: int) -> None:
    InputState.claimed_touches.add(touch_id)


def is_touch_claimed(touch_id: int) -> bool:
    return touch_id in InputState.claimed_touches


def unclaimed_taps() -> Iterator[Touch]:
    for touch in touches():
        if touch.started and not is_touch_claimed(touch.id):
            yield touch


def unclaimed_touches() -> Iterator[Touch]:
    for touch in touches():
        if not is_touch_claimed(touch.id):
            yield touch



from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.liveness import LivenessAnalysis, get_live
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, TempBlock

TEMP_SIZE = 4096


class AllocateBasic(CompilerPass):
    """Allocate temporary memory for temporary variables without considering lifetimes."""

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        offsets = {}
        index = 0

        def process(stmt):
            nonlocal index
            match stmt:
                case int():
                    return stmt
                case IRConst():
                    return stmt
                case IRPureInstr(op=op, args=args):
                    return IRPureInstr(
                        op=op,
                        args=[process(arg) for arg in args],
                    )
                case IRInstr(op=op, args=args):
                    return IRInstr(
                        op=op,
                        args=[process(arg) for arg in args],
                    )
                case IRGet(place=place):
                    return IRGet(place=process(place))
                case IRSet(place=place, value=value):
                    return IRSet(place=process(place), value=process(value))
                case BlockPlace() as place:
                    if isinstance(place.block, TempBlock):
                        if place.block not in offsets:
                            offsets[place.block] = index
                            index += place.block.size
                            if index >= TEMP_SIZE:
                                raise ValueError("Temporary memory limit exceeded")
                        return BlockPlace(10000, process(place.index), place.offset + offsets[place.block])
                    return BlockPlace(
                        process(place.block) if isinstance(place.block, BlockPlace) else place.block,
                        process(place.index),
                        process(place.offset),
                    )
                case _:
                    raise TypeError(f"Unsupported statement: {stmt}")

        for block in traverse_cfg_preorder(entry):
            block.statements = [process(statement) for statement in block.statements]
            block.test = process(block.test)

        return entry


class Allocate(CompilerPass):
    def requires(self) -> set[CompilerPass]:
        return {LivenessAnalysis()}

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        mapping = self.get_mapping(entry)
        for block in traverse_cfg_preorder(entry):
            updated_statements = [self.update_stmt(statement, mapping) for statement in block.statements]
            block.statements = [stmt for stmt in updated_statements if stmt is not None]
            block.test = self.update_stmt(block.test, mapping)
        return entry

    def update_stmt(self, stmt, mapping: dict[TempBlock, int]):
        match stmt:
            case int() | float() | IRConst():
                return stmt
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.update_stmt(arg, mapping) for arg in args])
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.update_stmt(arg, mapping) for arg in args])
            case IRGet(place=place):
                return IRGet(place=self.update_stmt(place, mapping))
            case IRSet(place=place, value=value):
                # Do some dead code elimination here which is pretty much free since we already have liveness analysis,
                # and prevents an error from the dead block place being missing from the mapping.
                live = get_live(stmt)
                is_live = not (
                    (isinstance(place, BlockPlace) and isinstance(place.block, TempBlock) and place.block not in live)
                    or (isinstance(value, IRGet) and place == value.place)
                )
                if is_live:
                    return IRSet(place=self.update_stmt(place, mapping), value=self.update_stmt(value, mapping))
                elif isinstance(value, IRInstr) and value.op.side_effects:
                    return self.update_stmt(value, mapping)
                else:
                    return None
            case BlockPlace(block=block, index=index, offset=offset):
                if isinstance(block, TempBlock):
                    if block.size == 0:
                        offset = -1
                    else:
                        assert block in mapping, f"Missing mapping for {block}, dead variable not removed?"
                        offset = mapping[block] + offset
                    return BlockPlace(block=10000, index=self.update_stmt(index, mapping), offset=offset)
                return BlockPlace(
                    block=self.update_stmt(block, mapping),
                    index=self.update_stmt(index, mapping),
                    offset=self.update_stmt(offset, mapping),
                )
            case _:
                raise TypeError(f"Unsupported statement: {stmt}")

    def get_mapping(self, entry: BasicBlock) -> dict[TempBlock, int]:
        interference = self.get_interference(entry)
        offsets: dict[TempBlock, int] = {}

        for block, others in sorted(interference.items(), key=lambda x: (-x[0].size, x[0].name)):
            size = block.size
            offset = 0
            for other in sorted(others, key=lambda x: offsets.get(x, 0) + x.size):
                if other not in offsets:
                    continue
                other_offset = offsets[other]
                if offset + size > other_offset or other_offset + other.size > offset:
                    offset = other_offset + other.size
            if offset + size > TEMP_SIZE:
                raise ValueError("Temporary memory limit exceeded")
            offsets[block] = offset

        return offsets

    def get_interference(self, entry: BasicBlock) -> dict[TempBlock, set[TempBlock]]:
        result = {}
        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                if not isinstance(stmt, IRSet):
                    continue
                live = {p for p in get_live(stmt) if isinstance(p, TempBlock) and p.size > 0}
                for place in live:
                    if place not in result:
                        result[place] = set(live)
                    else:
                        result[place].update(live)
        return result


class AllocateFast(Allocate):
    """A bit faster than Allocate but a bit less optimal."""

    def get_mapping(self, entry: BasicBlock) -> dict[TempBlock, int]:
        interference = self.get_interference(entry)
        offsets: dict[TempBlock, int] = dict.fromkeys(interference, 0)
        end_offsets: dict[TempBlock, int] = dict.fromkeys(interference, 0)

        for block, others in sorted(interference.items(), key=lambda x: (-x[0].size, x[0].name)):
            size = block.size
            offset = max((end_offsets[other] for other in others), default=0)
            if offset + size > TEMP_SIZE:
                raise ValueError("Temporary memory limit exceeded")
            offsets[block] = offset
            end_offsets[block] = offset + size

        return offsets



from dataclasses import dataclass


@dataclass
class CallbackInfo:
    name: str
    py_name: str
    supports_order: bool
    returns_value: bool


preprocess_callback = CallbackInfo(
    name="preprocess",
    py_name="preprocess",
    supports_order=True,
    returns_value=False,
)
spawn_order_callback = CallbackInfo(
    name="spawnOrder",
    py_name="spawn_order",
    supports_order=True,
    returns_value=True,
)
should_spawn_callback = CallbackInfo(
    name="shouldSpawn",
    py_name="should_spawn",
    supports_order=False,
    returns_value=True,
)
initialize_callback = CallbackInfo(
    name="initialize",
    py_name="initialize",
    supports_order=False,
    returns_value=False,
)
update_sequential_callback = CallbackInfo(
    name="updateSequential",
    py_name="update_sequential",
    supports_order=True,
    returns_value=False,
)
touch_callback = CallbackInfo(
    name="touch",
    py_name="touch",
    supports_order=True,
    returns_value=False,
)
update_parallel_callback = CallbackInfo(
    name="updateParallel",
    py_name="update_parallel",
    supports_order=False,
    returns_value=False,
)
terminate_callback = CallbackInfo(
    name="terminate",
    py_name="terminate",
    supports_order=False,
    returns_value=False,
)
spawn_time_callback = CallbackInfo(
    name="spawnTime",
    py_name="spawn_time",
    supports_order=False,
    returns_value=True,
)
despawn_time_callback = CallbackInfo(
    name="despawnTime",
    py_name="despawn_time",
    supports_order=False,
    returns_value=True,
)
update_spawn_callback = CallbackInfo(
    name="updateSpawn",
    py_name="update_spawn",
    supports_order=False,
    returns_value=True,
)
render_callback = CallbackInfo(
    name="render",
    py_name="render",
    supports_order=False,
    returns_value=False,
)
navigate_callback = CallbackInfo(
    name="navigate",
    py_name="navigate",
    supports_order=False,
    returns_value=False,
)
update_callback = CallbackInfo(
    name="update",
    py_name="update",
    supports_order=False,
    returns_value=False,
)


def _by_name(*callbacks: CallbackInfo) -> dict[str, CallbackInfo]:
    return {cb.py_name: cb for cb in callbacks}


PLAY_CALLBACKS = _by_name(
    preprocess_callback,
    spawn_order_callback,
    should_spawn_callback,
    initialize_callback,
    update_sequential_callback,
    touch_callback,
    update_parallel_callback,
    terminate_callback,
)
WATCH_ARCHETYPE_CALLBACKS = _by_name(
    preprocess_callback,
    spawn_time_callback,
    despawn_time_callback,
    initialize_callback,
    update_sequential_callback,
    update_parallel_callback,
    terminate_callback,
)
WATCH_GLOBAL_CALLBACKS = _by_name(
    update_spawn_callback,
)
PREVIEW_CALLBACKS = _by_name(
    preprocess_callback,
    render_callback,
)



from pydori.lib.stage import draw_stage
from pydori.tutorial.framework import current_phase_time, reset_phase, update_end, update_start
from pydori.tutorial.phases import PHASES
from sonolus.script.globals import level_memory


@level_memory
class TutorialState:
    current_phase: int


def inc_phase():
    TutorialState.current_phase += 1
    TutorialState.current_phase %= len(PHASES)
    reset_phase()


def dec_phase():
    TutorialState.current_phase -= 1
    TutorialState.current_phase %= len(PHASES)
    reset_phase()


def run_current_phase():
    for i, phase in enumerate(PHASES):
        if i == TutorialState.current_phase:
            is_done = phase(current_phase_time())
            if is_done:
                inc_phase()
            return


def update():
    update_start()
    draw_stage()
    run_current_phase()
    update_end()
